# coding: UTF-8
from __future__ import absolute_import, unicode_literals

import pytest

from barril import units


def CreateUnitDatabaseLenTime():
    """
    :rtype: UnitDatabase
    :returns:
        Returns a unit database with length and time quantity types
    """
    unit_database = units.UnitDatabase()
    unit_database.AddUnit("length", "milimeters", "mm", "%f * 1000.0", "%f / 1000.0")
    unit_database.AddUnit("length", "centimeters", "cm", "%f * 100.0", "%f / 100.0")
    unit_database.AddUnitBase("length", "meters", "m")
    unit_database.AddUnit("length", "kilometers", "km", "%f / 1000.0", "%f * 1000.0")
    unit_database.AddUnitBase("time", "seconds", "s")
    unit_database.AddUnit("time", "minutes", "minutes", "%f / 60.0", "%f * 60.0")

    unit_database.AddCategory(
        category="Table size",
        quantity_type="length",
        valid_units=["m", "cm"],
        default_unit="m",
    )
    unit_database.AddCategory(
        category="City size",
        quantity_type="length",
        valid_units=["km", "m"],
        default_unit="km",
    )
    unit_database.AddCategory(category="Time", quantity_type="time", default_unit="s")

    return unit_database


def CreateUnitDatabaseStartUnits():
    """
    :rtype: UnitDatabase
    :returns:
        Returns a unit database with length and time quantity types
    """
    unit_database = units.UnitDatabase()

    unit_database.AddUnitBase("length", "meters", "m")
    unit_database.AddUnit("length", "kilometers", "km", "%f / 1000.0", "%f * 1000.0")
    unit_database.AddUnit("length", "centimeters", "cm", "%f * 100.0", "%f / 100.0")
    unit_database.AddCategory("length", "length")

    return unit_database


def CreateUnitDatabaseWellLength():
    """
    :rtype: UnitDatabase
    :returns:
        Returns a unit database with length and time quantity types
    """
    unit_database = units.UnitDatabase()

    unit_database.AddUnit("length", "milimeters", "mm", "%f * 1000.0", "%f / 1000.0")
    unit_database.AddUnitBase("length", "meters", "m")
    unit_database.AddUnit("length", "centimeters", "cm", "%f * 100.0", "%f / 100.0")
    unit_database.AddUnit("length", "kilometers", "km", "%f / 1000.0", "%f * 1000.0")
    unit_database.AddUnit("length", "miles", "mi", "%f / 1609.347", "%f * 1609.347")
    unit_database.AddUnit("length", "inches", "in", "%f / 0.0254", "%f * 0.0254")
    unit_database.AddUnit(
        "length", "micrometers", "um", "%f * 1000000.0", "%f / 1000000.0"
    )
    unit_database.AddCategory("length", "length")
    unit_database.AddCategory("well-length", "length")
    unit_database.AddCategory("well-length-with-min-and-max", "length", min_value=0.0)

    return unit_database


def CreateUnitDatabasePoscLen(fill_categories=True):
    """
    :rtype: UnitDatabase
    :returns:
        Returns a unit database with "length" quantity type only

    :param bool fill_categories:
        If a category "length" should also be created
    """
    unit_database = units.UnitDatabase()
    unit_database.AddUnitBase("length", "meters", "m")
    unit_database.AddUnit("length", "kilometers", "km", "%f / 1000.0", "%f * 1000.0")

    if fill_categories:
        unit_database.AddCategory(category="length", quantity_type="length")
    return unit_database


def CreateUnitDatabaseLenPressure():
    """
    :rtype: UnitDatabase
    :returns:
        Returns a unit database with length and time quantity types
    """
    unit_database = units.UnitDatabase()

    unit_database.AddUnitBase("length", "meters", "m")
    unit_database.AddUnit("length", "kilometers", "km", "%f / 1000.0", "%f * 1000.0")
    unit_database.AddUnitBase("pressure", "pascal", "Pa")
    unit_database.AddUnit(
        "pressure", "pounds/square inch", "psi", "%f / 6894.757", "%f * 6894.757"
    )

    unit_database.AddCategory("length", "length")
    unit_database.AddCategory("pressure", "pressure")

    return unit_database


def CreateUnitDatabaseCustomConversion():
    from barril.units import UNKNOWN_QUANTITY_TYPE, UNKNOWN_UNIT

    unit_database = units.UnitDatabase()
    # add some units for testing
    unit_database.AddUnit(UNKNOWN_QUANTITY_TYPE, UNKNOWN_UNIT, UNKNOWN_UNIT, "%f", "%f")
    unit_database.AddUnit("length", "milimeters", "mm", "%f * 1000.0", "%f / 1000.0")
    unit_database.AddUnitBase("length", "meters", "m")
    unit_database.AddUnit("length", "centimeters", "cm", "%f * 100.0", "%f / 100.0")
    unit_database.AddUnit("length", "kilometers", "km", "%f / 1000.0", "%f * 1000.0")
    unit_database.AddUnit("length", "miles", "mi", "%f / 1609.347", "%f * 1609.347")
    unit_database.AddUnit("length", "inches", "in", "%f / 0.0254", "%f * 0.0254")
    unit_database.AddUnit(
        "length", "micrometers", "µm", "%f * 1000000.0", "%f / 1000000.0"
    )
    unit_database.AddUnitBase("temperature", "Celcius", "ºC")
    unit_database.AddCategory("length", "length", valid_units=["m", "mm", "cm", "km"])

    def C_to_F(x):
        return x * 1.8 + 32

    def F_to_C(x):
        return (x - 32) / 1.8

    unit_database.AddUnit("temperature", "Fahrenheit", "F", C_to_F, F_to_C)

    def C_to_K(x):
        return x + 273.15

    def K_to_C(x):
        return x - 273.15

    unit_database.AddUnit("temperature", "Kelvin", "K", C_to_K, K_to_C)

    return unit_database


@pytest.yield_fixture
def unit_database():
    """
    Fixture to be used whenever a test needs a clean UnitDatabase. When using this fixture, it's
    safe to call UnitDatabase.GetSingleton().
    """
    from barril.units.unit_database import UnitDatabase

    yield UnitDatabase.PushSingleton()
    UnitDatabase.PopSingleton()


@pytest.yield_fixture
def unit_database_posc():
    """
    Fixture to be used whenever a test needs a clean UnitDatabase. When using this fixture, it's
    safe to call UnitDatabase.GetSingleton().
    """
    unit_database = units.UnitDatabase()
    unit_database.FillUnitDatabaseWithPosc(unit_database)
    units.UnitDatabase.PushSingleton(unit_database)

    yield unit_database

    units.UnitDatabase.PopSingleton()


@pytest.fixture
def unit_database_empty():
    database = units.UnitDatabase()
    units.UnitDatabase.PushSingleton(database)

    yield database

    units.UnitDatabase.PopSingleton()


@pytest.fixture
def unit_database_len():
    database = units.UnitDatabase()
    units.UnitDatabase.PushSingleton(database)
    database.AddUnitBase("length", "meters", "m")
    database.AddUnit("length", "kilometers", "km", "%f / 1000.0", "%f * 1000.0")
    database.AddUnitBase("temperature", "degC", "degC")
    database.AddUnitBase("flow rate", "m3/s", "m3/s")
    database.AddCategory("length", "length")
    database.AddCategory("temperature", "temperature")
    database.AddCategory("flow rate", "flow rate", min_value=0.0)

    yield database

    units.UnitDatabase.PopSingleton()


@pytest.fixture
def unit_database_len_time():
    database = CreateUnitDatabaseLenTime()
    units.UnitDatabase.PushSingleton(database)

    yield database

    units.UnitDatabase.PopSingleton()


@pytest.fixture
def unit_database_start_units():
    database = CreateUnitDatabaseStartUnits()
    units.UnitDatabase.PushSingleton(database)

    yield database

    units.UnitDatabase.PopSingleton()


@pytest.fixture
def unit_database_well_length():
    database = CreateUnitDatabaseWellLength()
    units.UnitDatabase.PushSingleton(database)

    yield database

    units.UnitDatabase.PopSingleton()


@pytest.fixture
def unit_database_len_pressure():
    database = CreateUnitDatabaseLenPressure()
    units.UnitDatabase.PushSingleton(database)

    yield database

    units.UnitDatabase.PopSingleton()


@pytest.fixture
def unit_database_custom_conversion():
    database = CreateUnitDatabaseCustomConversion()
    units.UnitDatabase.PushSingleton(database)

    yield database

    units.UnitDatabase.PopSingleton()


@pytest.fixture
def unit_database_posc_len_no_category():
    database = CreateUnitDatabasePoscLen(fill_categories=False)
    units.UnitDatabase.PushSingleton(database)

    yield database

    units.UnitDatabase.PopSingleton()


@pytest.fixture
def unit_database_posc_len():
    database = CreateUnitDatabasePoscLen(fill_categories=True)
    units.UnitDatabase.PushSingleton(database)

    yield database

    units.UnitDatabase.PopSingleton()



from __future__ import unicode_literals

import six

# Custom cache for optimization purposes.
__bases_cache = {}


# ===================================================================================================
# AllBasesNames
# ===================================================================================================
def AllBasesNames(p_class):
    """
        :rtype: set with all the names of the bases classes of the given class.
    """
    try:
        return __bases_cache[p_class]
    except KeyError:
        result = set()
        for i_base in p_class.__bases__:
            result.add(i_base.__name__)
            result.update(AllBasesNames(i_base))
        return __bases_cache.setdefault(p_class, result)


# ===================================================================================================
# IsInstance
# ===================================================================================================
def IsInstance(p_object, p_class_name):
    """
    :param object p_object:
        The object we would like to test for.

    :param str p_class_name:
        Name or class to test if the object is an instance of.

    Like the built-in isinstance, but also accepts a class name as parameter.
    """
    try:
        # obtain the type of the class; using only type() is not enough, because some built-in
        # types don't respond to type() correctly (vtk objects for instance always return the
        # same type object)
        class_ = p_object.__class__
    except AttributeError:
        # some built-in objects don't have a __class__ attribute, but return its type correctly
        # from type()
        class_ = type(p_object)
    return IsSubclass(class_, p_class_name)


class _NotFoundSentinel(object):
    pass


# ===================================================================================================
# IsSubclass
# ===================================================================================================
def IsSubclass(p_class, p_class_name):
    """
    Like the built-in issubclass, but also accepts a class name as parameter.
    """
    is_instance = isinstance  # put it in locals

    if is_instance(p_class_name, six.string_types):
        if p_class_name == p_class.__name__:
            return True

        try:
            names = __bases_cache[p_class]
        except KeyError:
            return p_class_name in AllBasesNames(p_class)
        else:
            return p_class_name in names
    elif (
        is_instance(p_class_name, tuple)
        and len(p_class_name) > 0
        and is_instance(p_class_name[0], six.string_types)
    ):

        names = None

        for c in p_class_name:
            if c == p_class.__name__:
                return True

            if names is None:
                try:
                    names = __bases_cache[p_class]
                except KeyError:
                    names = AllBasesNames(p_class)

            if c in names:
                return True

        return False

    else:
        return issubclass(p_class, p_class_name)



# coding: utf-8
from __future__ import absolute_import, unicode_literals

import locale

from barril.basic.format_float import FormatFloat, MINUS_INFINITY, NAN, PLUS_INFINITY
from barril.units import Scalar


def testFormatFloat():
    """
    Convert unit of scalar where value is "0.0".
    """
    assert FormatFloat("%g", 0.0) == "0"
    assert FormatFloat("%g", -0.0) == "0"

    scalar = Scalar("length", 0.0, "m")
    assert FormatFloat("%g", scalar.GetValue()) == "0"

    assert locale.format_string("%g", scalar.GetValue("ft")) == "-0"
    assert FormatFloat("%g", scalar.GetValue("ft")) == "0"

    # Large float numbers on integer format.
    large_float_number = 1e+010 * 1.0
    assert FormatFloat("%d", large_float_number) == "10000000000"

    # Infinity
    assert FormatFloat("%.3g", PLUS_INFINITY) == "+INF"
    assert FormatFloat("%.3g", MINUS_INFINITY) == "-INF"
    assert FormatFloat("%.3g", NAN) == "-1.#IND"

    # Digit grouping
    assert FormatFloat("%.2f", 1234567, False) == "1234567.00"
    assert FormatFloat("%.2f", 1234567, True) == "1234567.00"



# coding: UTF-8
from __future__ import absolute_import, division, unicode_literals

import pytest
from pytest import approx

from six.moves import range

from barril import units
from barril.basic.format_float import FormatFloat
from barril.basic.fraction import FractionValue
from barril.units import (
    UNKNOWN_QUANTITY_TYPE,
    UNKNOWN_UNIT,
    Array,
    FixedArray,
    InvalidQuantityTypeError,
    InvalidUnitError,
    ObtainQuantity,
    Quantity,
    UnitsError,
)
from barril.units.unit_database import UnitDatabase


def testNotDefaultUnitDatabase():
    unit_database = units.UnitDatabase()
    with pytest.raises(AssertionError):
        unit_database.CheckDefaultUnitDatabase()


def testDefaultUnitDatabase(unit_database):
    unit_database.CheckDefaultUnitDatabase()


def testConv(unit_database_empty):
    unit_database = unit_database_empty

    # add some units for testing
    unit_database.AddUnitBase("temperature", "Celcius", "ºC")
    unit_database.AddUnit(
        "temperature", "Fahrenheit", "F", "%f * 1.8 + 32.0", " (%f - 32.0) / 1.8"
    )
    unit_database.AddUnit("temperature", "Kelvin", "K", "%f + 273.15", "%f - 273.15")
    unit_database.AddUnit(
        "temperature",
        "Rakini",
        "R",
        "%f * 1.8 + 32 + 459.67",
        "(%f - 32 - 459.67) / 1.8",
    )

    assert 50 == unit_database.Convert("temperature", "ºC", "F", 10)
    assert 10 == unit_database.Convert("temperature", "F", "ºC", 50)
    assert (
        approx(
            abs(-245.372222 - unit_database.Convert("temperature", "R", "ºC", 50)), 5
        )
        == 0
    )
    assert (
        approx(abs(581.67 - unit_database.Convert("temperature", "ºC", "R", 50)), 7)
        == 0
    )

    assert [50, 50, 50] == unit_database.Convert("temperature", "ºC", "F", [10, 10, 10])
    assert [10, 10, 10] == unit_database.Convert("temperature", "F", "ºC", [50, 50, 50])
    assert (10, 10, 10) == unit_database.Convert("temperature", "F", "ºC", (50, 50, 50))


def testPrecision():
    unit_database = units.UnitDatabase()
    unit_database.AddUnitBase("Compressibility", "1/Bars", "1/Bars")
    unit_database.AddUnit(
        "Compressibility",
        "1/Psi",
        "1/Psi",
        frombase="%f / 14.50377",
        tobase="%f * 14.50377",
    )
    unit_database.AddUnit(
        "Compressibility",
        "1/Atm",
        "1/Atm",
        frombase="%f / 0.986923",
        tobase="%f * 0.986923",
    )

    assert (
        approx(
            abs(
                14.50377 * 10e-6
                - unit_database.Convert("Compressibility", "1/Psi", "1/Bars", 1 * 10e-6)
            ),
            5,
        )
        == 0
    )
    assert (
        approx(
            abs(
                0.000063816588
                - unit_database.Convert("Compressibility", "1/Psi", "1/Bars", 4.4e-006)
            ),
            5,
        )
        == 0
    )


def testQuantityTypes(unit_database_custom_conversion):
    unit_database = unit_database_custom_conversion
    real = unit_database.GetQuantityTypes()
    real.sort()
    assert real == [UNKNOWN_QUANTITY_TYPE, "length", "temperature"]


def testUnitQuantityType(unit_database_custom_conversion):
    unit_database = unit_database_custom_conversion
    assert "length" == unit_database.GetQuantityType("m")


def testDistanceUnits(unit_database_custom_conversion):
    unit_database = unit_database_custom_conversion
    available_units = ["m", "mm", "cm", "km", "mi", "in", "µm"]
    real = unit_database.GetUnits("length")
    assert available_units == real


def testFindCase(unit_database_custom_conversion):
    unit_database = units.UnitDatabase()
    # add some units for testing
    unit_database.AddUnitBase("temperature", "mang1", "mA")
    unit_database.AddUnit("temperature", "mang2", "Ma", frombase="%f", tobase="%f")
    unit_database.AddUnit("temperature", "Celsius", "C", frombase="%f", tobase="%f")
    unit_database.AddUnit("temperature", "mang3", "MA", frombase="%f", tobase="%f")

    unit_database.AddCategory("my", "temperature")
    with pytest.raises(AssertionError):
        unit_database.FindUnitCase("my", "ma")
    assert "C" == unit_database.FindUnitCase("my", "c")
    assert "C" == unit_database.FindUnitCase("my", "C")


def testUnits(unit_database_custom_conversion):
    unit_database = unit_database_custom_conversion
    real = unit_database.GetUnits()
    real.sort()
    available_units = [
        UNKNOWN_UNIT,
        "m",
        "mm",
        "cm",
        "km",
        "mi",
        "in",
        "µm",
        "ºC",
        "F",
        "K",
    ]
    available_units.sort()
    assert available_units == real


def testBaseUnit(unit_database_custom_conversion):
    unit_database = unit_database_custom_conversion
    assert "m" == unit_database.GetBaseUnit("length")


def testUnitNames(unit_database_custom_conversion):
    unit_database = unit_database_custom_conversion
    names = set(
        [
            "meters",
            "milimeters",
            "centimeters",
            "kilometers",
            "miles",
            "inches",
            "micrometers",
        ]
    )
    onames = set(unit_database.GetUnitNames("length"))
    assert names == onames


def testClear(unit_database_custom_conversion):
    unit_database = unit_database_custom_conversion
    unit_database.Clear()
    with pytest.raises(units.InvalidQuantityTypeError):
        unit_database.CheckQuantityType("length")


def testCategory(unit_database_custom_conversion):
    unit_database = unit_database_custom_conversion
    unit_database.AddCategory(
        "my length",
        "length",
        ["mm", "m"],
        default_unit="m",
        default_value=15.5,
        min_value=-6e-10,
        max_value=2e5,
        is_min_exclusive=True,
        is_max_exclusive=False,
        caption="My Length",
    )

    assert unit_database.GetDefaultValue("my length") == 15.5
    assert unit_database.GetDefaultUnit("my length") == "m"
    assert sorted(list(unit_database.IterCategories())) == ["length", "my length"]

    quantity = ObtainQuantity("m", "my length")
    formatted_value = FormatFloat("%g", -6e-010)
    with pytest.raises(
        ValueError,
        match="Invalid value for My Length: %s. Must be > %s."
        % (formatted_value, formatted_value),
    ):
        quantity.CheckValue(-6e-10)

    quantity.CheckValue(0)  # without specifying unit
    mm_quantity = ObtainQuantity("mm", "my length")
    mm_quantity.CheckValue(2e5)

    with pytest.raises(
        ValueError, match="Invalid value for My Length: 200000. Must be <= 200000.0."
    ):
        mm_quantity.CheckValue(2e8 + 1)

    # Check the unit info using a category instead a quantity_type
    with pytest.raises(InvalidQuantityTypeError):
        unit_database.GetInfo("unknown length", "mm")

    unit_database.GetInfo("my length", "mm")
    unit_database.GetInfo(UNKNOWN_QUANTITY_TYPE, "I know it is m3", fix_unknown=True)


def testUniqueness(unit_database_custom_conversion):
    # trying to re-register milimeters (mm)
    unit_database = unit_database_custom_conversion
    with pytest.raises(RuntimeError):
        unit_database.AddUnit(
            "length", "milimeters", "mm", "%f * 1000.0", "%f / 1000.0"
        )


def testDefaultUnitWhenNoneIsPassed(unit_database_custom_conversion):
    unit_database = unit_database_custom_conversion
    unit_database.AddCategory("my category", "length", ["mm"], default_unit="mm")
    # Quantity without specifying unit, must use category's default unit and not raise
    default_when_none = ObtainQuantity(None, "my category")
    assert ObtainQuantity("mm", "my category") == default_when_none


def testDefaultUnitValid(unit_database_custom_conversion):
    """
    Test the situation where a category does not have the base unit among its valid units
    """
    unit_database = unit_database_custom_conversion
    unit_database.AddCategory("my category", "length", ["mm"])
    assert unit_database.GetCategoryInfo("my category").default_unit == "mm"


def testConvertionWithExponent(unit_database_custom_conversion):
    """
    Test conversions with different exponents.
    """
    unit_database = unit_database_custom_conversion
    assert (
        approx(
            abs(100 - unit_database.Convert("length", [("m", 1)], [("cm", 1)], 1)), 5
        )
        == 0
    )
    assert (
        approx(
            abs(10000 - unit_database.Convert("length", [("m", 2)], [("cm", 2)], 1)), 5
        )
        == 0
    )

    # Doesn't make sense changing the exponent in the from and to
    with pytest.raises(ValueError):
        unit_database.Convert("length", [("m", 2)], [("m", 1)], 1)


def testAddCategory(unit_database_custom_conversion):
    """
    Testing if AddCategory gives an error if trying to register twice.
    """
    unit_database = unit_database_custom_conversion
    unit_database.AddCategory(
        "my category", "length", valid_units=["mm", "m", "cm"], default_unit="mm"
    )
    with pytest.raises(UnitsError):
        unit_database.AddCategory("my category", "length", ["mm"], default_unit="mm")

    unit_database.AddCategory("derived category", "length")

    # Create a preferred unit set for categories
    assert unit_database.GetValidUnits("my category") == ["mm", "m", "cm"]
    assert unit_database.GetValidUnits("derived category") == ["m", "mm", "cm", "km"]


def testValidUnits(unit_database_custom_conversion):
    """
    Test while registering a category that passing None as valid_units returns the
    units of the quantity type.
    """
    unit_database = unit_database_custom_conversion
    unit_database.AddCategory("length", "length", override=True)
    info = unit_database.AddCategory("my category", "length", default_unit="mm")
    length_units = unit_database.GetUnits("length")
    assert info.valid_units is None
    assert unit_database.GetValidUnits("my category") == length_units


def testConvertQuantityTypeCheck(unit_database_custom_conversion):
    """
    Check for bug when trying to convert a unit to the same unit, but passing
    an invalid quantity type as argument.
    """
    unit_database = unit_database_custom_conversion
    with pytest.raises(InvalidQuantityTypeError):
        unit_database.Convert("XXX", "m", "mm", 100)
    with pytest.raises(InvalidUnitError):
        unit_database.Convert("length", "XXX", "m", 100)
    with pytest.raises(InvalidUnitError):
        unit_database.Convert("length", "m", "XXX", 100)


def testCheckValidUnits(unit_database_custom_conversion):
    """
    Make sure we check if the given valid_units are actually valid for that quantity_type
    """
    unit_database = unit_database_custom_conversion
    with pytest.raises(ValueError):
        unit_database.AddCategory(
            "my category", "length", default_unit="mm", valid_units="foooo"
        )


def testDiscoverCloseUnitMatches():
    unit_database = UnitDatabase.CreateDefaultSingleton()
    assert unit_database.FindSimilarUnitMatches("kg/day") == ["kg/d"]
    assert unit_database.FindSimilarUnitMatches("bbls/d") == ["bbl/d", "bbl/d2"]
    assert unit_database.FindSimilarUnitMatches("mg/l") == ["mg/L"]


def testDefaultCaption():
    unit_database = UnitDatabase.CreateDefaultSingleton()
    category_info = unit_database.GetCategoryInfo("angle per volume")
    assert category_info.caption == "Angle per Volume"


def testAddCategoryBasedOnCategory(unit_database_custom_conversion):
    unit_database = unit_database_custom_conversion
    unit_database.AddCategory(
        "my category",
        "length",
        valid_units=["mm", "m"],
        default_unit="mm",
        default_value=0.5,
    )

    unit_database.AddCategory(
        "my category 2", from_category="my category", valid_units=["mm"]
    )
    assert unit_database.GetValidUnits("my category 2") == ["mm"]

    unit_database.AddCategory(
        "my category 3", from_category="my category", default_unit="m"
    )
    assert unit_database.GetDefaultUnit("my category 3") == "m"

    unit_database.AddCategory(
        "my category 4", from_category="my category", default_value=0.8
    )
    assert unit_database.GetDefaultValue("my category 4") == 0.8
    assert unit_database.GetValidUnits("my category 4") == ["mm", "m"]


def testConvertFractionValues(unit_database_custom_conversion):
    """
    Test unit database handling fraction values.
    """
    db = unit_database_custom_conversion

    # Converting half meter to cm
    value = FractionValue(0, (1, 2))

    value = db.Convert("length", "m", "cm", value)
    assert float(value) == 50


def testRegisterTwoFunctionsForTheSameClass(unit_database_custom_conversion):
    """
    Test the behavior when we attempt to register two convert functions for the same class
    """

    def ConvertFunction1(*args, **kwargs):
        return 0

    db = unit_database_custom_conversion
    # There is a function specialized in convert fraction values. Attempting to register another
    # function should raise an error

    value = FractionValue(0, (1, 2))
    db.Convert("length", "m", "cm", value)

    with pytest.raises(AssertionError):
        db.RegisterAdditionalConversionType(FractionValue, ConvertFunction1)


def testNumpyConversion(unit_database_custom_conversion):
    unit_database = unit_database_custom_conversion

    from barril.units.posc import MakeBaseToCustomary, MakeCustomaryToBase

    f_unit_to_base = MakeCustomaryToBase(273.15, 1, 1, 0)
    f_base_to_unit = MakeBaseToCustomary(273.15, 1, 1, 0)
    unit_database.AddUnit(
        "My Temperature",
        "degrees Celsius",
        "degC",
        f_base_to_unit,
        f_unit_to_base,
        default_category=None,
    )
    f_unit_to_base = MakeCustomaryToBase(2298.35, 5, 9, 0)
    f_base_to_unit = MakeBaseToCustomary(2298.35, 5, 9, 0)
    unit_database.AddUnit(
        "My Temperature",
        "degree Fahrenheit",
        "degF",
        f_base_to_unit,
        f_unit_to_base,
        default_category=None,
    )
    f_unit_to_base = MakeCustomaryToBase(0, 1, 1, 0)
    f_base_to_unit = MakeBaseToCustomary(0, 1, 1, 0)
    unit_database.AddUnit(
        "My Temperature",
        "my Kelvin",
        "myK",
        f_base_to_unit,
        f_unit_to_base,
        default_category=None,
    )

    unit_database.AddCategory(
        "My Temperature",
        quantity_type="My Temperature",
        valid_units=["myK", "degC", "degF"],
        default_unit="myK",
        default_value=2,
        min_value=0,
        max_value=15,
        is_min_exclusive=False,
        is_max_exclusive=False,
        caption="My Temperature",
    )

    import numpy

    values = numpy.array(list(range(10, 13)), numpy.float32)
    arr = Array("My Temperature", values, "degC")
    obtained = arr.GetValues("degF")

    expected = [49.99995041, 51.79994965, 53.5999527]
    assert approx(obtained) == expected


def testNumpyWithMinMax(unit_database_custom_conversion):
    unit_database = unit_database_custom_conversion
    unit_database.AddCategory(
        "my length",
        "length",
        ["mm", "m"],
        default_unit="m",
        default_value=2,
        min_value=0,
        max_value=15,
        is_min_exclusive=False,
        is_max_exclusive=False,
        caption="My Length",
    )
    import numpy

    values = numpy.array(list(range(10)), numpy.float32)
    Array("my length", values, "mm")

    values = numpy.array(list(range(20000, 20010)), numpy.float32)
    # Raise validation error
    another = Array("my length", values, "mm")
    assert not another.IsValid()
    with pytest.raises(ValueError):
        another.CheckValidity()

    # Don't raise validation error if validate = False.
    arr = Array("my length", values, "mm")
    assert not arr.IsValid()

    another = arr.CreateCopy(values=values, unit="mm")
    assert not another.IsValid()

    # Same checks on FixedArray.
    values = numpy.array(list(range(10)), numpy.float32)
    FixedArray(len(values), "my length", values, "mm")

    values = numpy.array(list(range(20000, 20010)), numpy.float32)
    # Raise validation error
    another = FixedArray(len(values), "my length", values, "mm")
    assert not another.IsValid()

    # Don't raise validation error if validate = False.
    arr = FixedArray(len(values), "my length", values, "mm")
    assert not arr.IsValid()

    another = arr.CreateCopy(values=values, unit="mm")
    assert not another.IsValid()


def testInvalidUnitCategoryDoesntGenerateError(unit_database_custom_conversion):
    unit_database = unit_database_custom_conversion
    with pytest.raises(ValueError):
        unit_database.AddCategory(
            "my invalid length",
            "length",
            caption="Invalid length",
            # This is not a valid unit
            default_unit="Mm",
            min_value=0,
            max_value=99999.0,
            is_min_exclusive=False,
            is_max_exclusive=False,
            valid_units=["mm", "m"],
        )


def testGetValidUnitsEmptyQuantity(unit_database_custom_conversion):
    """
    Test that retrieving the valid units for the category of the empty quantity (as created
    by Quantity.CreateEmpty() correctly returns an empty list.
    """
    unit_database = unit_database_custom_conversion
    quantity = Quantity.CreateEmpty()
    assert unit_database.GetValidUnits(quantity.GetCategory()) == []



from __future__ import unicode_literals

import six
from six import raise_from

# Inspired by http://www.thescripts.com/forum/thread46361.html

if six.PY2:

    # ===============================================================================================
    # Reraise
    # ===============================================================================================
    def Reraise(exception, message, separator="\n"):
        """
        Raised the same exception given, with an additional message.

        :param Exception exception:
            Original exception being raised with additional messages

        :param str message:
            Message to be added to the given exception

        :param str separator:
            String separating `message` from the `exception`'s original message.

        e.g.
            try:
                raise RuntimeError('original message')
            except Exception, e:
                Reraise(e, 'message')

            >>> RuntimeError:
            >>> message
            >>> original message

            try:
                raise RuntimeError('original message')
            except Exception, e:
                Reraise(e, '[message]', separator=' ')

            >>> RuntimeError:
            >>> [message] original message
        """
        from barril._foundation.exceptions import ExceptionToUnicode

        # IMPORTANT: Do NOT use try/except mechanisms in this method or the sys.exc_info()[-1] will be invalid
        if hasattr(exception, "reraised_message"):
            current_message = exception.reraised_message
        else:
            current_message = ExceptionToUnicode(exception)

        # Build the new message
        if not current_message.startswith(separator):
            current_message = separator + current_message
        message = "\n" + message + current_message

        if exception.__class__ in _SPECIAL_EXCEPTION_MAP:
            # Handling for special case, some exceptions have different behaviors.
            exception = _SPECIAL_EXCEPTION_MAP[exception.__class__](*exception.args)

        elif exception.__class__ not in _SPECIAL_EXCEPTION_MAP.values():
            # In Python 2.5 overriding the exception "__str__" has no effect in "unicode()". Instead, we
            # must change the "args" attribute which is used to build the string representation.
            # Even though the documentation says "args" will be deprecated, it uses its first argument
            # in unicode() implementation and not "message".
            exception.args = (message,)

        exception.message = message
        # keep the already decoded message in the object in case this exception is reraised again
        exception.reraised_message = message

        # taken from source code of future.utils.raise_with_traceback
        import sys  # noqa

        exec("raise exception, None, sys.exc_info()[-1]")

    # ===================================================================================================
    # SPECIAL_EXCEPTIONS
    # ===================================================================================================
    class ReraisedKeyError(KeyError):
        def __init__(self, *args):
            KeyError.__init__(self, *args)
            self.message = None

        def __str__(self):
            return self.message

    class ReraisedOSError(OSError):
        def __init__(self, *args):
            OSError.__init__(self, *args)
            self.message = None

        def __str__(self):
            return self.message

    class ReraisedSyntaxError(SyntaxError):
        def __init__(self, *args):
            SyntaxError.__init__(self, *args)
            self.message = None

        def __str__(self):
            return self.message

    class ReraisedUnicodeDecodeError(UnicodeDecodeError):
        def __init__(self, *args):
            UnicodeDecodeError.__init__(self, *args)
            self.message = None

        def __str__(self):
            return self.message

    class ReraisedUnicodeEncodeError(UnicodeEncodeError):
        def __init__(self, *args):
            UnicodeEncodeError.__init__(self, *args)
            self.message = None

        def __str__(self):
            return self.message

    _SPECIAL_EXCEPTION_MAP = {
        KeyError: ReraisedKeyError,
        OSError: ReraisedOSError,
        SyntaxError: ReraisedSyntaxError,
        UnicodeDecodeError: ReraisedUnicodeDecodeError,
        UnicodeEncodeError: ReraisedUnicodeEncodeError,
    }
    # [[[end]]] (checksum: 87ea5d69d51083d7009b216f50cc2be5)
else:

    def Reraise(exception, message, separator="\n"):
        """
        Forwards to a `raise exc from cause` statement. Kept alive for backwards compatibility
        (`separator` argument only kept for this reason).
        """
        # Important: Don't create a local variable for the new exception otherwise we'll get a
        # cyclic reference between the exception and its traceback, meaning the traceback will
        # keep all frames (and their contents) alive.
        raise_from(type(exception)(message), exception)



from __future__ import unicode_literals

from barril._foundation.odict import odict


def testInsert():
    d = odict()
    d[1] = "alpha"
    d[3] = "charlie"

    assert list(d.items()) == [(1, "alpha"), (3, "charlie")]

    d.insert(0, 0, "ZERO")
    assert list(d.items()) == [(0, "ZERO"), (1, "alpha"), (3, "charlie")]

    d.insert(2, 2, "bravo")
    assert list(d.items()) == [(0, "ZERO"), (1, "alpha"), (2, "bravo"), (3, "charlie")]

    d.insert(99, 4, "echo")
    assert list(d.items()) == [
        (0, "ZERO"),
        (1, "alpha"),
        (2, "bravo"),
        (3, "charlie"),
        (4, "echo"),
    ]


def testDelWithSlices():
    d = odict()
    d[1] = 1
    d[2] = 2
    d[3] = 3

    del d[1:]

    assert len(d) == 1
    assert d[1] == 1



from __future__ import unicode_literals

import threading


# ===================================================================================================
# SingletonError
# ===================================================================================================
class SingletonError(RuntimeError):
    """
    Base class for all Singleton-related exceptions.
    """


# ===================================================================================================
# SingletonAlreadySetError
# ===================================================================================================
class SingletonAlreadySetError(SingletonError):
    """
    Trying to set a singleton when the class already have one defined.
    """


# ===================================================================================================
# SingletonNotSetError
# ===================================================================================================
class SingletonNotSetError(SingletonError):
    """
    Trying to clear a singleton when there's none defined.
    """


# ===================================================================================================
# PushPopSingletonError
# ===================================================================================================
class PushPopSingletonError(SingletonError):
    """
    Trying to set a singleton between a PushSingleton/PopSingleton calls.
    """


# ===================================================================================================
# Singleton
# ===================================================================================================
class Singleton(object):
    """
    Base class for singletons.

    A Singleton class should have a unique instance during the lifetime of the application. Besides
    the functionality of obtaining the singleton instance, this class also provides methods to push
    and pop singletons, useful for testing, where you push a singleton into a known state during
    setUp and pops it back during tearDown
    """

    # name of the attribute that holds the stack of singletons
    __singleton_stack_start_index = 0
    __lock = threading.RLock()

    _singleton_classes = set()

    @staticmethod
    def ResetDefaultSingletonInstances():
        """
        This singleton class is intended to be used in tests with the push / pop protocol. However some singleton
        dependencies might be hidden away from the test creator (or even be introduced after the test creation) making
        easy for a code to access and change the default class singleton (for example registering on its callbacks).

        This code is intended to clear any change made in such default singletons. Pushed singletons will not be cleared
        because if a test has correctly pushed it singleton, it is reasonable to assume that the test will correctly
        clean (pop) it.

        TODO: ETK-1235 As soon as the classes with ResetInstance are moved to do not be a singleton, then this method
        can be removed.
        """
        for cls in Singleton._singleton_classes:
            if cls._UsingDefaultSingleton():
                instance = cls.GetSingleton()
                instance.ResetInstance()

    @classmethod
    def GetSingleton(cls):
        """
        :rtype: Singleton
        :returns:
            Returns the current singleton instance.

        .. note:: This function is thread-safe, but all the other methods (such as SetSingleton,
            PushSingleton, PopSingleton, etc) are not (which should be Ok as those are mostly
            test-related, as singletons shouldn't really be changed after the application is up
            especially on multi-threaded environments).
        """
        Singleton._singleton_classes.add(cls)

        try:
            # Make common case faster.
            return cls.__singleton_singleton_stack__[-1]
        except (AttributeError, IndexError):
            with cls.__lock:
                # Only lock if the 'fast path' did not work.
                stack = cls._ObtainStack()

                if not stack:  # Faster than doing len(stack) == 0
                    return cls.SetSingleton(None)

                return stack[-1]

    @classmethod
    def SetSingleton(cls, instance):
        """
        Sets the current singleton.

        :param Singleton instance:
            The Singleton to pass as parameter

        :rtype: Singleton
        :returns:
            The singleton passed as parameter.

        @raise PushPopSingletonError
        @raise SingletonAlreadySetError
        """
        stack = cls._ObtainStack()

        # Error if we trying to use SetSingleton between a Push/Pop
        if len(stack) != cls.__singleton_stack_start_index:
            raise PushPopSingletonError(
                "SetSingleton can not be called between a Push/Pop pair."
            )

        if len(stack) > 0:
            raise SingletonAlreadySetError(
                "SetSingleton can only be called when there is no singleton set."
            )

        # Obtain default instance (if needed)
        if instance is None:
            instance = cls.CreateDefaultSingleton()

        # Set the stack[0] as the singleton
        if len(stack) == 0:
            stack.append(instance)
            cls.__singleton_stack_start_index = 1
        else:
            stack[0] = instance

        assert cls.__singleton_stack_start_index == 1

        return instance

    @classmethod
    def _UsingDefaultSingleton(cls):
        """
        Checks if the current singleton instance is the default instance.

        :rtype: bool
        :returns:
            True if the current singleton instance is the default created instance. Returns False if
            the current instance is a pushed singleton or if no instance is currently set
        """
        stack = cls._ObtainStack()
        has_pushed = len(stack) != cls.__singleton_stack_start_index
        has_singleton = cls.HasSingleton()

        return has_singleton and not has_pushed

    def ResetInstance(self):
        """
        Restore the instance original configuration. Singleton classes should not have a internal state to reset
        (as described in issue ETK-1235), so subclasses that implement this method are strong candidates to
        be refactored to do not be a singleton.

        This method is used to avoid interference between tests while ETK-1235 is not implemented.
        """
        pass

    @classmethod
    def ClearSingleton(cls):
        """
        Clears the current singleton
        """
        stack = cls._ObtainStack()

        # Error if we trying to use ClearSingleton between a Push/Pop
        if len(stack) != cls.__singleton_stack_start_index:
            raise PushPopSingletonError(
                "ClearSingleton can not be called between a Push/Pop pair."
            )

        if not stack:
            raise SingletonNotSetError(
                "ClearSingleton can only be called when THERE IS singleton set."
            )

        del stack[0]
        cls.__singleton_stack_start_index = 0

    @classmethod
    def HasSingleton(cls):
        """
        Do we have any singleton set?

        :rtype: bool
        :returns:
            True if there's a singleton set.
        """
        stack = cls._ObtainStack()
        return len(stack) > 0

    @classmethod
    def CreateDefaultSingleton(cls):
        """
        Creates the default singleton instance, that will be used when no singleton has been installed.
        By default, tries to create the class without constructor.

        :rtype: Singleton
        :returns:
            an instance of the singleton subclass
        """
        return cls()

    # Push/Pop -------------------------------------------------------------------------------------

    @classmethod
    def PushSingleton(cls, instance=None):
        """
        Pushes the given singleton to the top of the stack. The previous singleton will be restored
        when PopSingleton is called.

        :param Singleton instance:
            The singleton to install as the current one. If not given, a new singleton default
            is created.

        :rtype: Singleton
        :returns:
            The current singleton.
        """
        if instance is None:
            instance = cls.CreateDefaultSingleton()
        stack = cls._ObtainStack()

        stack.append(instance)
        return instance

    @classmethod
    def PopSingleton(cls):
        """
        Restores the singleton that was the current before the last PushSingleton.

        :rtype: Singleton
        :returns:
            Return the removed singleton.
        """
        stack = cls._ObtainStack()

        if len(stack) == cls.__singleton_stack_start_index:
            raise PushPopSingletonError(
                "PopSingleton called without a pair PushSingleton call"
            )

        return cls._ObtainStack().pop(-1)

    @classmethod
    def _ObtainStack(cls):
        """
        Obtains the stack of singletons.

        :rtype: list
        :returns:
            The singleton stack.
        """
        try:
            return cls.__singleton_singleton_stack__
        except AttributeError:
            assert (
                cls is not Singleton
            ), "This method can only be called from a Singleton subclass."
            stack = []
            cls.__singleton_singleton_stack__ = stack
            return stack

    @classmethod
    def GetStackCount(cls):
        """
        @return int:
            The number of elements added int the stack using PushSingleton.
        """
        stack = cls._ObtainStack()
        return len(stack) - cls.__singleton_stack_start_index



from __future__ import absolute_import, unicode_literals

from ._array import Array
from ._quantity import Quantity

__all__ = [str("FixedArray")]  # pylint: disable=invalid-all-object


# ===================================================================================================
# FixedArray
# ===================================================================================================
class FixedArray(Array):
    """Represents an Array with fixed number of elements.
    """

    _dimension = None

    def __init__(self, dimension, category, values=None, unit=None):
        """
        :param int dimension:
            The dimension for this array.

        :type category: string or IQuantity
        :param category:
            The category for the unit point or the IQuantity for this object
            (in this case, the unit will be ignored (if it is passed)).

        :type values: sequence or numpy array
        :param values:
            A sequence with its initial values.

        :param unicode unit:
            Unit (not used if a quantity is passed).
        """
        if dimension < 2:
            raise ValueError("Dimension MUST be 2 or more")
        self._dimension = dimension

        Array.__init__(self, category, values, unit)

    def _InternalCreateWithQuantity(
        self, quantity, values=None, unit_database=None, dimension=None, value=None
    ):
        if value is not None:
            if values is not None:
                raise ValueError("Duplicated values parameter given")

            values = value

        if dimension is None:
            try:
                if self._dimension is None:
                    self._dimension = len(values)
            except AttributeError:
                pass

            dimension = self._dimension

        elif hasattr(self, "_dimension"):
            if self._dimension is not None and dimension != self._dimension:
                raise ValueError(
                    "Dimension re-definition mismatch: %s != %s"
                    % (self._dimension, dimension)
                )

        if dimension < 2:
            raise ValueError("Dimension MUST be 2 or more")
        self._dimension = dimension

        if values is None:
            values = [0.0] * dimension
        self.CheckValues(values, dimension)

        Array._InternalCreateWithQuantity(self, quantity, values)

    def CreateCopy(self, values=None, unit=None, category=None, **kwargs):
        return Array.CreateCopy(
            self,
            values=values,
            unit=unit,
            category=category,
            dimension=self._dimension,
            **kwargs
        )

    # Values ---------------------------------------------------------------------------------------
    def _GetDefaultValue(self, category_info, unit=None):
        """

        :param category_info:
        :param unit:
        """
        return [0.0] * self._dimension

    def CheckValues(self, values, dimension=None):
        """Checks whether the dimensions consistent with the dimensions in this unit point
        """
        if dimension is None:
            dimension = self.dimension

        if len(values) != dimension:
            msg = "Values must have %d elements, but has %d"
            raise ValueError(msg % (dimension, len(values)))

    @classmethod
    def CreateEmptyArray(cls, dimension, values=None):
        """
        Allows the creation of a array that does not have any associated category nor unit.

        :rtype: Array
        :returns:
            Returns an empty array.
        """
        if values is None:
            values = [0.0] * dimension

        quantity = Quantity.CreateEmpty()
        return cls.CreateWithQuantity(quantity, dimension=dimension, values=values)

    # Dimension ------------------------------------------------------------------------------------
    def GetDimension(self):
        return self._dimension

    dimension = property(GetDimension)

    # Equality -------------------------------------------------------------------------------------
    def __eq__(self, other):
        return Array.__eq__(self, other) and self.dimension == other.dimension

    def __reduce__(self):
        """
        Defining reduce so that we can pickle fixed arrays.
        """
        return (
            FixedArray,
            (
                self._dimension,
                self._quantity,
                self.values,
                None,  # Unit defined in quantity
            ),
        )

    def ChangingIndex(self, index, value, use_value_unit=True):
        """
        Creates a FixedArray from based on this FixedArray changing a single value based on the
        passed index.

        i.e.: array.ChangingIndex(0, Scalar(1.0, 'm'))
              will create a new FixedArray where the index == 0 is changed to the passed value.

        :param Scalar|float|tuple value:
            The value to be used to set at the given index.

        :param int index:
            The index which should be changed.

        :param bool use_value_unit:
            If True and a Scalar is passed, the newly created array will have the unit of the
            scalar, not of the original array, otherwise (if False) the FixedArray unit will be
            kept.

        :return FixedArray:
            The created FixedArray.
        """
        from barril.units import Scalar
        from barril.units import FixedArray

        if isinstance(value, tuple):
            scalar = Scalar(self.GetValues()[index], self.GetUnit()).CreateCopy(*value)

        elif not isinstance(value, Scalar):
            scalar = Scalar(value, self.GetUnit())

        else:
            scalar = value

        if use_value_unit:
            quantity = scalar.GetQuantity()
        else:
            quantity = self.GetQuantity()

        values = list(self.GetValues(quantity.GetUnit()))
        values[index] = scalar.GetValue(quantity.GetUnit())
        return FixedArray(self.dimension, quantity, tuple(values))

    def IndexAsScalar(self, index, quantity=None):
        """
        :param int index:
            The index which should be gotten as a Scalar.

        :param IQuantity quantity:
            The quantity in which we want the Scalar (uses the FixedArray quantity
            if not passed).

        :return Scalar:
            A Scalar representation of the given index.
        """
        from barril.units import Scalar

        if quantity is None:
            quantity = self.GetQuantity()
        return Scalar(quantity, self.GetValues(unit=quantity.GetUnit())[index])



from __future__ import absolute_import, unicode_literals

from barril.units import Scalar
from barril.units.scalar_validation.scalar_min_max_validator import (
    ScalarMinMaxValidator
)


def _CreateTestCategories(db):
    db.AddCategory(
        category="test category",
        quantity_type="dimensionless",
        override=True,
        default_unit="-",
        min_value=1.0,
        max_value=50.0,
        valid_units="-",
        is_min_exclusive=False,
        is_max_exclusive=False,
    )

    db.AddCategory(
        category="category exclusive",
        quantity_type="dimensionless",
        override=True,
        default_unit="-",
        min_value=1.0,
        max_value=50.0,
        valid_units="-",
        is_min_exclusive=True,
        is_max_exclusive=True,
        default_value=5.0,
    )


def testScalarValidationMsgs(unit_database):
    def _Check(scalar, value, unit, expected_msg):
        some_scalar = scalar.CreateCopy(value=value, unit=unit)
        obtained_msg = ScalarMinMaxValidator.CreateScalarCheckErrorMsg(
            some_scalar, "Some Property"
        )
        assert obtained_msg == expected_msg

    _CreateTestCategories(unit_database)

    some_scalar = Scalar("test category", 10.0, "-")

    # Test value below minimum -----------------------------------------------------------------
    expected_error_msg = "Error in Some Property. Invalid value for Test Category: 0. Must be greater or equal to 1.0."
    _Check(some_scalar, 0.0, "-", expected_error_msg)

    # Test value above maximum -----------------------------------------------------------------
    expected_error_msg = "Error in Some Property. Invalid value for Test Category: 51. Must be less or equal to 50.0."
    _Check(some_scalar, 51.0, "-", expected_error_msg)

    # Test no error without exclusive ----------------------------------------------------------
    _Check(some_scalar, 1.0, "-", None)
    _Check(some_scalar, 50.0, "-", None)

    # Test using min and max exclusive ---------------------------------------------------------
    some_scalar = Scalar("category exclusive", 10.0, "-")

    # Test value below minimum -----------------------------------------------------------------
    expected_error_msg = "Error in Some Property. Invalid value for Category Exclusive: 1. Must be greater than 1.0."
    _Check(some_scalar, 1.0, "-", expected_error_msg)

    # Test value above maximum -----------------------------------------------------------------
    expected_error_msg = "Error in Some Property. Invalid value for Category Exclusive: 50. Must be less than 50.0."
    _Check(some_scalar, 50.0, "-", expected_error_msg)

    # Test no error with exclusive -------------------------------------------------------------
    _Check(some_scalar, 49.0, "-", None)
    _Check(some_scalar, 2.0, "-", None)



from __future__ import unicode_literals

import collections

import six


class _OrderedDict(collections.OrderedDict):
    def insert(self, index, key, value, dict_setitem=dict.__setitem__):
        """
        Convenience method to have same interface as `ruamel.ordereddict`, which as traditionally
        used on Python 2.
        """
        self[key] = value
        # Determine which direction is cheaper to move items first. If new item is more to the left
        # of center, move items to its left to first, otherwise it is cheaper to move items to
        # right to last.
        #
        # Note that `move_to_end` is a O(1) operation that just swaps endpoints of underlying
        # double linked list maintained by C-extension ordered dict.
        if (len(self) - index) <= (len(self) // 2):
            moved = [k for i, k in enumerate(self.keys()) if i >= index and k != key]
            last = True
        else:
            moved = reversed(
                [k for i, k in enumerate(self.keys()) if i < index or k == key]
            )
            last = False
        for k in moved:
            self.move_to_end(k, last=last)

    def __delitem__(self, key):
        if isinstance(key, slice):
            # Properly deal with slices (based on order).
            keys = list(self.keys())
            for k in keys[key]:
                collections.OrderedDict.__delitem__(self, k)

        else:
            collections.OrderedDict.__delitem__(self, key)


if six.PY2:
    from _ordereddict import ordereddict as odict  # pylint: disable=unused-import
else:
    odict = _OrderedDict  # pylint: disable=unused-import



from __future__ import unicode_literals

import pytest
import six
from six.moves import range  # @UnresolvedImport

from barril._foundation.types_ import (
    AsList,
    Boolean,
    CheckBasicType,
    CheckEnum,
    CheckFormatString,
    CheckIsNumber,
    CheckType,
    Flatten,
    Intersection,
    IsBasicType,
    IsNumber,
    MergeDictsRecursively,
    OrderedIntersection,
    StructMap,
    _GetKnownNumberTypes,
)


def testBoolean():
    assert Boolean("TRUE")
    assert Boolean("true")
    assert Boolean("yes")
    assert Boolean("1")
    assert not Boolean("false")
    assert not Boolean("no")
    assert not Boolean("0")

    with pytest.raises(ValueError):
        Boolean("INVALID")


def testPassing():
    class Foo(object):
        pass

    CheckType(Foo(), Foo)
    CheckType(Foo(), (int, Foo))
    CheckType(99, (int, Foo))


def testRaising():
    with pytest.raises(TypeError):
        CheckType("hellou", int)

    with pytest.raises(TypeError):
        CheckType("hellou", (int, float))

    with pytest.raises(TypeError):
        CheckType(99, (six.text_type, float))


def testCheckFormatString():
    CheckFormatString("%s", 1)
    CheckFormatString("%s m", 1)

    with pytest.raises(ValueError):
        CheckFormatString("%s m %s", 1)

    with pytest.raises(ValueError):
        CheckFormatString("%s m %s", 1, 3, 3)


def testIfCustomMessageIsAppendedToDefaultMessage():
    message = "Zero is not unicode!"

    with pytest.raises(TypeError) as exception:
        CheckType(0, six.text_type, message)

    assert message in six.text_type(exception.value)


def testBasicType():
    class NonBasic(object):
        ""

    assert IsBasicType(1)
    assert not IsBasicType([1])
    assert IsBasicType([1], accept_compound=True)
    assert IsBasicType({1: [1]}, accept_compound=True)
    assert IsBasicType({1: set([1])}, accept_compound=True)
    assert IsBasicType(frozenset([1, 2]), accept_compound=True)
    assert IsBasicType([1, [2, [3]]], accept_compound=True)
    assert not IsBasicType({1: NonBasic()}, accept_compound=True)
    assert not IsBasicType({NonBasic(): 1}, accept_compound=True)

    assert not IsBasicType(NonBasic(), accept_compound=True)
    assert not IsBasicType([NonBasic()], accept_compound=True)
    assert not IsBasicType([1, [NonBasic()]], accept_compound=True)

    assert CheckBasicType(0)
    with pytest.raises(TypeError):
        CheckBasicType([0])


def testCheckEnum():
    for i in range(10):
        CheckEnum(i, list(range(10)))

    with pytest.raises(ValueError):
        CheckEnum(-1, list(range(10)))
    with pytest.raises(ValueError):
        CheckEnum(11, list(range(10)))
    with pytest.raises(ValueError):
        CheckEnum("foo", list(range(10)))


def testCheckNumber():
    numpy = pytest.importorskip("numpy")

    for number_class in [float] + list(six.integer_types):
        converted = number_class(1)
        assert IsNumber(converted)

    # Checking numpy numbers
    collection = numpy.zeros(1, numpy.float32)
    assert IsNumber(collection[0])


def testGetKnownNumberTypes(monkeypatch):
    import sys

    numpy = pytest.importorskip("numpy")

    expected = {float, complex, numpy.number}
    expected.update(set(six.integer_types))
    assert set(_GetKnownNumberTypes()) == expected

    monkeypatch.setitem(sys.modules, "numpy", None)
    expected.remove(numpy.number)
    assert set(_GetKnownNumberTypes()) == expected


def testCheckIsNumber():
    assert CheckIsNumber(1)
    assert CheckIsNumber(1.)
    if six.PY2:
        assert CheckIsNumber(long(1))  # noqa
    with pytest.raises(TypeError):
        CheckIsNumber("alpha")


class ListWithoutIter(object):
    def __init__(self, *args, **kwargs):
        self.contents = []
        for item in args:
            self.contents.append(item)

    def __getitem__(self, index):
        return self.contents[index]


def testAsList():
    a = [1, 2, 3]
    assert a is AsList(a)
    assert ["a"] == AsList("a")
    assert ["a"] == AsList(("a",))
    assert ["a"] == AsList(set("a"))


def testFlatten():
    a = [[[1], [2]], [3]]
    assert Flatten(a) == [1, 2, 3]

    a = [1, 2, [3, 4], (5, 6)]
    assert Flatten(a) == [1, 2, 3, 4, 5, 6]


def testFlattenOnClassWithoutIter():
    a = ListWithoutIter(ListWithoutIter(0, 1), 2, 3)
    assert Flatten(a) == [0, 1, 2, 3]


def testFlattenOnClassWithoutIterForStrings():
    a = ListWithoutIter(ListWithoutIter("a", "bb"), "ccc")
    assert Flatten(a) == ["a", "bb", "ccc"]


def testFlattenForStrings():
    a = [["a", "bb"], "ccc"]
    assert Flatten(a) == ["a", "bb", "ccc"]


def testFlattenForUnicodeStrings():
    a = [["a", "bb"], "ccc"]
    assert Flatten(a) == ["a", "bb", "ccc"]


def testFlattenForTuples():
    a = [(0, "a"), (1, "b"), ((2, "c"), 42)]
    assert Flatten(a) == [0, "a", 1, "b", 2, "c", 42]


def testFlattenSkipSpecificClass():
    obj = ListWithoutIter("a", "b")
    a = [obj, "c", ["d", "e"]]
    assert Flatten(a, skip_types=[ListWithoutIter]) == [obj, "c", "d", "e"]


def testFlattenSkipTypeOfSubclass():
    class Foo(ListWithoutIter):
        def __init__(self, *args, **kwargs):
            super(Foo, self).__init__(*args, **kwargs)

    obj = Foo()
    a = [obj, "c", "d"]
    assert Flatten(a, skip_types=[ListWithoutIter]) == [obj, "c", "d"]


def testMergeDictsRecursively():
    dict_1 = {"a": 1, "b": 2}
    dict_2 = {"c": 3, "d": 4}
    assert MergeDictsRecursively(dict_1, dict_2) == {"a": 1, "b": 2, "c": 3, "d": 4}


def testMergeDictsRecursivelyDictsOnTheRightHaveHigherPrecedence():
    dict_1 = {"a": 1, "b": 2}
    dict_2 = {"b": 3, "d": 4}
    assert MergeDictsRecursively(dict_1, dict_2) == {"a": 1, "b": 3, "d": 4}


def testMergeDictsRecursivelyManyLevelsOfRecursion():
    dict_1 = {
        "a": 0,
        "b": {"replaced_inner_b": 0, "kept_inner_b": 0},
        "c": {"inner_c": {"replaced_inner_c": 0, "kept_inner_c": 0}},
    }
    dict_2 = {
        "b": {"replaced_inner_b": 42, "added_inner_b": 0},
        "c": {"inner_c": {"replaced_inner_c": 42, "added_inner_c": 0}},
    }
    assert MergeDictsRecursively(dict_1, dict_2) == {
        "a": 0,
        "b": {"kept_inner_b": 0, "replaced_inner_b": 42, "added_inner_b": 0},
        "c": {
            "inner_c": {"replaced_inner_c": 42, "added_inner_c": 0, "kept_inner_c": 0}
        },
    }


def testMergeDictsRecursivelyWhenLeftHasDictKeyButRightDoesnt():
    dict_1 = {"a": {"inner_a": 0}}
    dict_2 = {"a": 42}
    assert MergeDictsRecursively(dict_1, dict_2) == {"a": 42}


def testMergeDictsRecursivelyWhenRightHasDictKeyButLeftDoesnt():
    dict_1 = {"a": 42}
    dict_2 = {"a": {"inner_a": 0}}
    assert MergeDictsRecursively(dict_1, dict_2) == {"a": {"inner_a": 0}}


def testMergeDictsWithWrongTypes():
    with pytest.raises(AttributeError):
        MergeDictsRecursively("Foo", {})

    with pytest.raises(TypeError):
        MergeDictsRecursively({}, "Foo")

    with pytest.raises(TypeError) as excinfo:
        MergeDictsRecursively({}, "Foo")

    if six.PY2:
        expected = (
            'Wrong types passed. Expecting two dictionaries, got: "dict" and "unicode"'
        )
    else:
        expected = (
            'Wrong types passed. Expecting two dictionaries, got: "dict" and "str"'
        )
    assert six.text_type(excinfo.value) == expected


def testIntersection():
    alpha = [3, 2, 1]
    bravo = [2, 3, 4]
    assert Intersection(alpha, bravo) == {2, 3}

    assert Intersection() == set()


def testOrderedIntersection():
    alpha = [1, 2, 3]
    bravo = [2, 3, 4]
    assert OrderedIntersection(alpha, bravo) == [2, 3]

    alpha = [3, 1, 2]
    bravo = [2, 3, 4]
    assert OrderedIntersection(alpha, bravo) == [3, 2]

    alpha = [1, 2, 3]
    bravo = [4, 5, 6]
    assert OrderedIntersection(alpha, bravo) == []


def testStructMap():

    a = {"alpha": [1, 2, 3], "bravo": (1, 2, 3)}

    obtained = StructMap(
        a, func=six.text_type, conditional=lambda x: isinstance(x, int)
    )
    expected = {"alpha": ["1", "2", "3"], "bravo": ("1", "2", "3")}
    assert obtained == expected



from __future__ import absolute_import, unicode_literals

import six

from barril.units import ObtainQuantity


# ===================================================================================================
# ScalarMinMaxValidator
# ===================================================================================================
class ScalarMinMaxValidator(object):
    """
    Simple helper class to create default checkers.

    It implements methods to create simple messages of scalar value limits checking
    """

    @classmethod
    def CreateScalarCheckWarningMsg(cls, scalar, name):
        """
        :param Scalar scalar:
            The scalar to be checked against its limits
        :param unicode name:
            The scalar property name

        :returns unicode:
            The built warning message saying if the scalar is less or greater than its limits
        """
        predicate = cls._ScalarCheckMsgPredicate(scalar)

        if predicate is not None:
            return "Warning in %s. %s" % (name, predicate)

        return None

    @classmethod
    def CreateScalarCheckErrorMsg(cls, scalar, name):
        """
         :param Scalar scalar:
            The scalar to be checked against its limits
        :param unicode name:
            The scalar property name

        :returns unicode:
            The built error message saying if the scalar is less or greater than its limits
        """
        predicate = cls._ScalarCheckMsgPredicate(scalar)

        if predicate is not None:
            return "Error in %s. %s" % (name, predicate)

        return None

    @classmethod
    def _ScalarCheckMsgPredicate(cls, scalar):
        """
        :param Scalar scalar:
            The scalar to be checked against its limits

        :returns unicode:
            The built message saying if the scalar is less or greater than its limits
        """
        try:
            quantity = ObtainQuantity(scalar.GetUnit(), scalar.GetCategory())
            quantity.CheckValue(scalar.GetValue(), use_literals=True)

        except ValueError as error:
            return six.text_type(error)

        return None



from __future__ import absolute_import, division, unicode_literals

from fractions import Fraction as StdFraction
from functools import total_ordering

import six

SMALL = 1e-8

NumberType = (six.integer_types, float)
SequenceType = (tuple, list)


@total_ordering
class Fraction(object):
    """
    Numerator, denominator similar to standard fraction with support to
    float values in numerator and denominator which will be automatically
    reduced to a simpler fraction. Class has also an special handling for
    rounding and representation issues for numbers.
    """

    def __init__(self, a, b=None):
        if a == float("inf") or a == float("-inf"):
            raise ValueError("Numerator cannot be an infinite value")

        if b == float("inf") or b == float("-inf"):
            raise ValueError("Denominator cannot be an infinite value")

        assert isinstance(a, NumberType) and (
            (isinstance(b, NumberType) and b != 0) or b is None
        ), (a, b)
        if b is None:
            b = 1
        elif b < 0:
            a, b = -a, -b

        while abs(a - round(a)) > SMALL:
            a *= 10
            b *= 10
        a = round(a)

        if isinstance(a, float) or isinstance(b, float):
            a = StdFraction(a)
            b = StdFraction(b)
            self.x = a / b
        else:
            self.x = StdFraction(a, b)

    def __len__(self):
        return 2

    def __getitem__(self, key):
        x = tuple([self.numerator, self.denominator])
        return x[key]

    def __setitem__(self, key, value):
        assert isinstance(value, NumberType) and (value or key != 1), value
        x = list([self.numerator, self.denominator])
        x[key] = value
        self.x = StdFraction(*x)

    def __str__(self):
        from barril.basic.format_float import FormatFloat

        return (
            FormatFloat("%g", self.numerator)
            + "/"
            + FormatFloat("%g", self.denominator)
        )

    if six.PY2:
        __unicode__ = __str__
        del __str__

    def __repr__(self):
        return repr(self.x)

    def __float__(self):
        return float(self.x)

    def __add__(self, other):
        if isinstance(other, NumberType):
            other = Fraction(other)
        x = self.x + other.x
        return Fraction(x.numerator, x.denominator)

    def __radd__(self, other):
        return self + other

    def __neg__(self):
        x = -self.x
        return Fraction(x.numerator, x.denominator)

    def __sub__(self, other):
        return self + (-other)

    def __rsub__(self, other):
        return -(self - other)

    def __mul__(self, other):
        if classify(other) == -1:
            return other * self  # hope a list structure can sort itself out
        if isinstance(other, NumberType):
            other = Fraction(other)
        x = Fraction(
            self.numerator * other.numerator, self.denominator * other.denominator
        )
        x.reduce()
        return x

    def __rmul__(self, other):
        if classify(other) == -1:
            raise ValueError(self, other)
        return self * other

    def __truediv__(self, other):
        if isinstance(other, NumberType):
            other = Fraction(other)
        return self * other.inv()

    def __rtruediv__(self, other):
        return self.inv() * other

    def __pow__(self, other):
        if abs(other - other) < SMALL:
            if other < 0:
                return Fraction(self.denominator ** -other, self.numerator ** -other)
            else:
                return Fraction(self.numerator ** other, self.denominator ** other)
        else:
            return Fraction(float(self) ** other)

    def __old_cmp__(self, other):
        if other == float("inf"):
            return -1
        if other == float("-inf"):
            return 1

        if isinstance(other, NumberType):
            other = Fraction(other)
        t = self.numerator * other.denominator - other.numerator * self.denominator
        if t < 0:
            return -1
        if t > 0:
            return 1
        return 0

    def __eq__(self, other):
        return self.__old_cmp__(other) == 0

    if six.PY2:

        def __ne__(self, other):
            return not self == other

    def __lt__(self, other):
        return self.__old_cmp__(other) == -1

    def __abs__(self):
        return Fraction(abs(self.numerator), self.denominator)

    def __mod__(self, other):
        if isinstance(other, NumberType):
            other = Fraction(other)
        x = self.x % other.x
        return Fraction(x.numerator, x.denominator)

    def reduce(self):
        "Express the fraction in it's lowest form.  Changes the object, no return."
        self.x = StdFraction(self.numerator, self.denominator)

    def inv(self):
        "return 1/self as a fraction"
        x = 1 / self.x
        return Fraction(x.numerator, x.denominator)

    def copy(self):
        "Returns a copy of the fraction"
        return Fraction(self.numerator, self.denominator)

    def get_numerator(self):
        """
        Returns the numerator of the fraction.

        :rtype: C{float}
        :returns:
            the numerator
        """
        return self.x.numerator

    def set_numerator(self, numerator):
        """
        Sets the numerator of the fraction.

        :type numerator: C{float}
        :param numerator:
            the numerator
        """
        if numerator == float("inf") or numerator == float("-inf"):
            raise ValueError("Numerator cannot be an infinite value")

        if isinstance(numerator, float):
            a = Fraction(numerator)
            b = Fraction(self.denominator)
            self.x = a.x / b.x
        else:
            self.x = StdFraction(numerator, self.denominator)

    numerator = property(get_numerator, set_numerator)

    def get_denominator(self):
        """
        Returns the denominator of the fraction.

        :rtype: C{float}
        :returns:
            the numerator
        """
        return self.x.denominator

    def set_denominator(self, denominator):
        """
        Sets the denominator of the fraction.

        :type denominator: C{float}
        :param denominator:
            the numerator
        """
        if denominator == float("inf") or denominator == float("-inf"):
            raise ValueError("Denominator cannot be an infinite value")

        if isinstance(denominator, float):
            a = Fraction(self.numerator)
            b = Fraction(denominator)
            self.x = a.x / b.x
        else:
            self.x = StdFraction(self.numerator, denominator)

    denominator = property(get_denominator, set_denominator)


def classify(instance):
    "Decide if instance is a sequence or a number. Returns 1 for number, -1 for sequence"
    if isinstance(instance, NumberType) or isinstance(instance, Fraction):
        return 1
    if isinstance(instance, SequenceType):
        return -1
    try:
        iter(instance)
        return -1
    except TypeError:
        return 1



from __future__ import unicode_literals

import locale

import six

# ===================================================================================================
# ExceptionToUnicode
# ===================================================================================================
if six.PY2:

    def ExceptionToUnicode(exception):
        """
        Obtains unicode representation of an Exception.

        This wrapper is used to circumvent Python 2.7 problems with built-in exceptions with unicode
        messages.

        Steps used:
            * Try to obtain Exception.__unicode__
            * If the exception message is a unicode, returns it.
            * Try to obtain Exception.__str__ and decode with utf-8
            * Try to obtain Exception.__str__ and decode with locale.getpreferredencoding
            * If all fails, return Exception.__str__ and decode with (ascii, errors='replace')

        :param Exception exception:

        :return six.text_type:
            Unicode representation of an Exception.
        """

        try:
            # First, try to obtain __unicode__ as defined by the Exception
            return six.text_type(exception)
        except UnicodeError:
            pass

        # When message is a unicode, just return it
        import warnings

        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            # exception.message deprecated since Python 2.6
            exception_message = getattr(exception, "message", None)
        if isinstance(exception_message, unicode):  # noqa
            return exception_message

        bytes_msg = six.binary_type(exception)
        try:
            # If that fails, try decoding with utf-8 which is the strictest and will complain loudly.
            return bytes_msg.decode("utf-8")
        except UnicodeError:
            pass
        try:
            # If that fails, try decoding with locale
            return bytes_msg.decode(locale.getpreferredencoding())
        except UnicodeError:
            pass

        # If all failed, give up and decode with ascii replacing errors.
        return bytes_msg.decode(errors="replace")


else:

    def ExceptionToUnicode(exception):
        """
        Python 3 exception handling already deals with string and six.binary_type (and mixed) error messages. Here we
        will only append the original exception message to the returned message (this is automatically done in Python 2
        since the original exception message is added into the new exception while Python 3 keeps the original exception
        as a separated attribute
        """
        messages = []
        while exception:
            messages.append(str(exception))
            exception = exception.__cause__ or exception.__context__
        return "\n".join(messages)



from __future__ import absolute_import, unicode_literals

import six

from barril._foundation.types_ import IsNumber
from barril.basic.format_float import FormatFloat
from barril.units.unit_database import UnitDatabase

from ._abstractvaluewithquantity import AbstractValueWithQuantityObject
from ._quantity import Quantity

__all__ = [str("Array")]  # pylint: disable=invalid-all-object


class Array(AbstractValueWithQuantityObject):
    """
    Array represents a sequence of values that also have an unit associated.

    Some ways to construct it (note that usually numpy arrays should be used).

    .. code-block:: python

        Array(numpy.array([0, 1, 2, 3, 4], numpy.float64), 'm')

        Array([0, 1, 2, 3, 4], 'm')

        Array('length', [0, 1, 2, 3, 4], 'm')

        Array(ObtainQuantity('m', 'length'), [0, 1, 2, 3, 4])
    """

    def __init__(self, category, values=None, unit=None):
        AbstractValueWithQuantityObject.__init__(
            self, category, value=values, unit=unit
        )

    def _InternalCreateWithQuantity(
        self, quantity, values=None, unit_database=None, value=None
    ):
        if value is not None:
            if values is not None:
                raise ValueError("Duplicated values parameter given")
            values = value

        self._value = values
        self._unit_database = unit_database or UnitDatabase.GetSingleton()
        self._quantity = quantity
        self._is_valid = None
        self._validity_exception = None

    def CheckValidity(self):
        """
        :raises ValueError: when current value is wrong somehow (out of limits, for example).
        """
        self.ValidateValues(self._value, self._quantity)

    def CreateCopy(self, values=None, unit=None, category=None, **kwargs):
        return AbstractValueWithQuantityObject.CreateCopy(
            self, value=values, unit=unit, category=category, **kwargs
        )

    # Values ---------------------------------------------------------------------------------------
    def GetAbstractValue(self, unit=None):
        """
        :param str unit: this is the unit in which we want the values
        :rtype: list(number)
        :returns:
            the values stored. May be an a list of int, float, etc.
        """
        values = self._value
        if unit is None or unit == self._quantity._unit:
            return values

        def IsListOfTuples(v):
            try:
                return len(v) > 0 and isinstance(v[0], tuple)
            except TypeError:
                return (
                    False
                )  # numpy raises a TypeError if it's a 0D array, so ignores it

        if IsListOfTuples(values):
            result = []
            Convert = self._quantity.Convert
            for elem in values:
                result.append(tuple(Convert(v, unit) for v in elem))
            return values.__class__(result)

        else:
            return self._quantity.Convert(values, unit)

    GetValues = GetAbstractValue
    values = property(GetAbstractValue)

    def _GetDefaultValue(self, category_info, unit=None):
        """

        :param category_info:
        :param unit:
        """
        return []

    def ValidateValues(self, values, quantity):
        """Set the value to store in this values_quantity. May be an int,
        float, numarray, list of floats, etc.

        :type values: sequence(values) or numpy array.
        :param values:
            The values to be set.

        :param unicode unit:
            The unit of the values being passed (note that GetUnit will still return the previous
            unit set -- this unit is only to indicate the internal value).
        """
        if self._is_valid is True:
            return

        if self._validity_exception is not None:
            raise self._validity_exception

        try:
            self._DoValidateValues(values, quantity)
        except Exception as e:
            self._validity_exception = e
            self._is_valid = False
            raise
        else:
            self._is_valid = True

    def _DoValidateValues(self, values, quantity):
        """
        .. seealso:: :meth:`.ValidateValues`
        """
        is_derived = quantity.IsDerived()
        if not is_derived:
            # only check min, max if we have only 1 category (otherwise, we won't have a valid assumption
            # about the actual values)
            category_info = quantity.GetCategoryInfo()

            if (
                category_info.min_value is not None
                or category_info.max_value is not None
            ):
                # verify if values are in the given limits (if needed)
                CheckValue = quantity.CheckValue
                if len(values) > 0:
                    if isinstance(values[0], tuple):
                        for value in values:
                            if isinstance(value, tuple):
                                for v in value:
                                    CheckValue(v)
                    else:
                        import numpy

                        isnam = numpy.isnan

                        # Search for the first non-NaN value to initialize MIN/MAX.
                        is_numpy = isinstance(values, numpy.ndarray)

                        iterator = iter(values)

                        for value in iterator:
                            if isnam(value):
                                continue

                            # Iterate until we get a non-nan number
                            min_value = max_value = value

                            # Keep on the iteration now that we can already make the check.
                            for value in iterator:
                                if isnam(value):
                                    # NaNs would fail the min_value validation below.
                                    continue

                                if value < min_value:
                                    min_value = value

                                elif value > max_value:
                                    max_value = value

                            if is_numpy:
                                min_value = float(min_value)
                                max_value = float(max_value)

                            CheckValue(min_value)
                            CheckValue(max_value)

                            # Break the outer 'for' used just to get the min/max
                            break

    @classmethod
    def CreateEmptyArray(cls, values=None):
        """
            Allows the creation of a array that does not have any associated
            category nor unit.

            :rtype: Array
        """
        if values is None:
            values = []

        quantity = Quantity.CreateEmpty()
        return cls.CreateWithQuantity(quantity, values=values)

    # Equality -------------------------------------------------------------------------------------
    def __eq__(self, other):
        if not isinstance(other, Array):
            return False

        return (
            tuple(self.values) == tuple(other.values)
            and self._quantity == other._quantity
            and self.unit == other.unit
        )

    def __repr__(self):
        values_str = "[%s]" % ", ".join(six.text_type(v) for v in self.values)
        return "%s(%s, %s, %s)" % (
            self.__class__.__name__,
            self.GetQuantityType(),
            values_str,
            self.GetUnit(),
        )

    def __str__(self):
        """
        Should return a user-friendly representation of this object.

        :rtype: unicode
        :returns:
            The formatted string
        """
        if len(self.values) > 0 and isinstance(self.values[0], tuple):
            values_str = " ".join(six.text_type(v) for v in self.values)
        else:
            values_str = " ".join((FormatFloat("%g", v)) for v in self.values)

        return values_str + self.GetFormattedSuffix()

    if six.PY2:
        __unicode__ = __str__
        del __str__

    # Basic operators ------------------------------------------------------------------------------
    def __len__(self):
        return len(self.values)

    def __getitem__(self, index):
        return self.values[index]

    def __iter__(self):
        return iter(self.values)

    def __truediv__(self, other):
        return self._DoOperation(self, other, "Divide")

    def __mul__(self, other):
        return self._DoOperation(self, other, "Multiply")

    def __add__(self, other):
        return self._DoOperation(self, other, "Sum")

    def __sub__(self, other):
        return self._DoOperation(self, other, "Subtract")

    # Right-Basic operators ------------------------------------------------------------------------
    def __rdiv__(self, other):
        return self._DoOperation(other, self, "Divide")

    def __rmul__(self, other):
        return self._DoOperation(other, self, "Multiply")

    def __radd__(self, other):
        return self._DoOperation(other, self, "Sum")

    def __rsub__(self, other):
        return self._DoOperation(other, self, "Subtract")

    def _DoOperation(self, p1, p2, operation):
        """
            Actually go on and do an operation considering the data we have to transform considering
            any combination of: number, list and numpy
        """
        from ._value_generator import _ValueGenerator
        import numpy

        # get the quantities and setup the value generator properly
        if IsNumber(p1) or isinstance(p1, numpy.ndarray):
            values_iteration = _ValueGenerator(p1, p2.values)
            q2 = p2.GetQuantity()
            q1 = Quantity.CreateEmpty()

        elif IsNumber(p2) or isinstance(p2, numpy.ndarray):
            values_iteration = _ValueGenerator(p1.values, p2)
            q1 = p1.GetQuantity()
            q2 = Quantity.CreateEmpty()

        else:
            values_iteration = _ValueGenerator(p1.values, p2.values)
            q1 = p1.GetQuantity()
            q2 = p2.GetQuantity()

        unit_database = self.GetUnitDatabase()
        operation = getattr(unit_database, operation)

        # if handling numpy, just call it all at once!
        if values_iteration.IsNumpy():
            v0, v1 = next(iter(values_iteration))
            q, v = operation(q1, q2, v0, v1)
            return self.__class__.CreateWithQuantity(q, v)
        else:
            # not numpy: create a new structure to hold the values
            result = []
            for v0, v1 in values_iteration:
                q, v = operation(q1, q2, v0, v1)
                result.append(v)

            if values_iteration.IsTuple():
                result = tuple(result)
            return self.__class__.CreateWithQuantity(q, result)



"""
This module provides the implementation of an Quantity object.
"""
# NoQuantityReplacement
from __future__ import absolute_import, unicode_literals

import six
from six.moves import range, zip  # @UnresolvedImport

from barril._foundation.odict import odict
from barril.units.unit_database import UnitsError
from ._unit_constants import UNKNOWN_UNIT
from .unit_database import UnitDatabase

__all__ = [str("Quantity")]  # pylint: disable=invalid-all-object


# ===================================================================================================
# _ObtainReduced
# ===================================================================================================
def _ObtainReduced(state):
    """
    :param list state:
        The value returned from _Quantity.__reduce__ (used in pickle protocol).
    """
    unknown_unit_caption = state.pop(-1)
    return ObtainQuantity(odict(state), None, unknown_unit_caption)


# ===================================================================================================
# ObtainQuantity
# ===================================================================================================
def ObtainQuantity(unit, category=None, unknown_unit_caption=None):
    """
    :type unit: unicode or odict(unicode -> list(unicode, int))
    :param unit:
        Either the string representing the unit or an ordered dict with the composing unit
        information (if composing all the info, including the category will be received in this
        parameter).

    :param unicode category:
        The category for the quantity. If not given it's gotten based on the unit passed.

    :param unicode unknown_unit_caption:
        The caption for the unit (used if unknown).

    :rtype Quantity:
    """
    unit_database = UnitDatabase.GetSingleton()
    quantities_cache = unit_database.quantities_cache

    if unit.__class__ in (list, tuple):
        # It may be a derived unit with list(tuple(unicode, int)) -- in which case the category
        # must also be a list (of the same size)
        if len(unit) == 1 and unit[0][1] == 1:
            # Although passed as composing, it's a simple case
            unit = unit[0][0]
            if category.__class__ in (list, tuple):
                category = category[0]
        else:
            assert category.__class__ in (list, tuple)
            unit = odict(
                (cat, unit_and_exp) for (cat, unit_and_exp) in zip(category, unit)
            )
            category = None

    if hasattr(unit, "items"):
        assert category is None
        if len(unit) == 1 and next(six.itervalues(unit))[1] == 1:
            # Although passed as composing, it's a simple case
            category, (unit, _exp) = next(six.iteritems(unit))
        else:
            key = tuple(
                (category, tuple(unit_and_exp))
                for (category, unit_and_exp) in six.iteritems(unit)
            )
            if unknown_unit_caption:
                key += (unknown_unit_caption,)
            try:
                return quantities_cache[key]
            except KeyError:
                quantity = quantities_cache[key] = _Quantity(
                    unit, None, unknown_unit_caption
                )
                return quantity

    key = (category, unit, unknown_unit_caption)
    try:
        return quantities_cache[key]
    except KeyError:
        pass  # Just go on with the regular flow.

    assert unit.__class__ != bytes, "unit must be given as unicode string always"
    if unit.__class__ != six.text_type:
        if category is None:
            raise AssertionError("Currently only supporting unit as a string.")
        else:
            # Unit is given by the category
            unit = unit_database.GetDefaultUnit(category)
        quantity = quantities_cache[key] = _Quantity(
            category, unit, unknown_unit_caption
        )
        return quantity

    elif category is None:
        category = unit_database.GetDefaultCategory(unit)
        if not category:
            raise UnitsError("Unable to get default category for: %s" % (unit,))

        key_with_resolved_category = (category, unit, unknown_unit_caption)
        try:
            return quantities_cache[key_with_resolved_category]
        except KeyError:
            quantity = quantities_cache[key_with_resolved_category] = _Quantity(
                category, unit, unknown_unit_caption
            )
            # Cache it with None category too.
            quantities_cache[key] = quantity
            return quantity

    else:
        quantities_cache[key] = quantity = _Quantity(
            category, unit, unknown_unit_caption
        )
        return quantity


# ===================================================================================================
# ReadOnlyError
# ===================================================================================================
class ReadOnlyError(NotImplementedError):
    """
    Error thrown if some change is attempted in the quantity (as it's now read-only).
    """


# ===================================================================================================
# Quantity
# ===================================================================================================
class Quantity(object):
    """
    .. note:: This class has nothing but factory methods. The real Quantity implementation is at
    _Quantity (which is what's returned from ObtainQuantity).

    .. note:: The preferred way to get a Quantity is through ObtainQuantity, although
    Quantity(category, unit) was maintained for backward-compatibility.
    """

    __slots__ = []

    def __new__(cls, category, unit, unknown_unit_caption=None):
        """
        Backward-compatible way of obtaining a quantity through the 'default' constructor.

        Redirects values to ObtainQuantity.
        .. see:: ObtainQuantity for parameters.
        """
        return ObtainQuantity(unit, category, unknown_unit_caption)

    _EMPTY_QUANTITY = None

    @classmethod
    def CreateEmpty(cls):
        """
        Create a quantity without any internal unit.

        :rtype: Quantity
        :returns:
            Returns an empty quantity.
        """
        empty = cls._EMPTY_QUANTITY
        if empty is None:
            empty = cls._EMPTY_QUANTITY = ObtainQuantity(odict())
        return empty

    @classmethod
    def CreateDerived(
        cls, category_to_unit_and_exps, resulting_class=None, unknown_unit_caption=None
    ):
        """
        Same as _CreateDerived, but always validating the category and units.

        .. see:: _CreateDerived for parameters.
        """
        return cls._CreateDerived(
            category_to_unit_and_exps,
            resulting_class,
            unknown_unit_caption=unknown_unit_caption,
        )

    @classmethod
    def _CreateDerived(
        cls,
        category_to_unit_and_exps,
        resulting_class=None,
        validate_category_and_units=True,
        unknown_unit_caption=None,
    ):
        """
        Create a category that represents a derived quantity (usually resulting from operations
        among quantities).

        :type category_to_unit_and_exps: odict(unicode->(list(unicode, int)))
        :param category_to_unit_and_exps:
            This odict defines the category as well as the unit in a way that we can specify exponents.

        :param type resulting_class:
            Deprecated (no longer used).

        :param bool validate_category_and_units:
            If True, the category and units will be validated (otherwise, it'll just create it
            without trying to validate any value (note that it's meant to be used internally
            to make creating the quantity faster without having to actually validate it)

        :rtype: cls
        :returns:
            An instance that represents a quantity with the given categories, units and exponents.
        """
        unit_database = UnitDatabase.GetSingleton()

        if validate_category_and_units:
            assert hasattr(
                category_to_unit_and_exps, "items"
            ), "validate_category_and_units needs to be a dict"
            for category, (unit, _exp) in six.iteritems(category_to_unit_and_exps):
                # will do the checkings needed for validation (category/unit)
                # so that we can later just store the internal information without
                # any other check.

                # changes to accomodate making operations with derived units (internally the
                # information must be a dict)
                if category.__class__ != six.text_type:
                    raise TypeError(
                        "Only unicode is accepted. %s is not." % category.__class__
                    )

                # Getting the category should be enough to know that it's valid.
                category_info = unit_database.GetCategoryInfo(category)

                # if no unit was given, assume base unit for this quantity_type
                if unit is None:
                    unit = category_info.default_unit
                else:
                    if unit.__class__ != six.text_type:
                        raise TypeError(
                            "Only unicode is accepted. %s is not." % unit.__class__
                        )
                    unit_database.CheckQuantityTypeUnit(
                        category_info.quantity_type, unit
                    )

        return ObtainQuantity(
            odict(
                (category, unit_and_exp[:])
                for (category, unit_and_exp) in six.iteritems(category_to_unit_and_exps)
            ),
            unknown_unit_caption=unknown_unit_caption,
        )


# ===================================================================================================
# Quantity
# ===================================================================================================
class _Quantity(Quantity):
    """
    The quantity is an object that has its associated category, quantity type and unit.

    Internally we represent the information as a dict with the information needed for derived
    units in the following way:

    category -> [unit,exp]

    Note that in this way we may have the same unit appearing in different places (so, the expoent
    may not be gotten directly only from one place, it must be checked against its multiple places)
    """

    __slots__ = [
        "__weakref__",
        "_category",
        "_category_info",
        "_category_to_unit_and_exps",
        "_composing_categories",
        "_composing_units",
        "_composing_units_joining_exponents",
        "_hash",
        "_is_derived",
        "_quantity_type",
        "_tobase",
        "_unit",
        "_unit_database",
        "_unknown_unit_caption",
        "_category_info",
        "_configured",
    ]

    def __new__(cls, category, unit, unknown_unit_caption=None):
        """
        Overridden because we don't want to call the Quantity.__new__ (which would result
        in a StackOverflowError).
        """
        return object.__new__(cls)

    def __copy__(self):
        """
        As we're now immutable, always return itself.
        """
        return self

    def __deepcopy__(self, *args, **kwargs):
        """
        As we're now immutable, always return itself.
        """
        return self

    def __reduce__(self):
        """
        Used in pickle protocol

        :rtype: tuple(callable, args)
        :returns:
            Used during the pickle so that we can restore it as the same instance it was before
            (i.e.: don't create a new instance).
        """
        lst = list(
            (category, unit_and_exp)
            for (category, unit_and_exp) in six.iteritems(
                self._category_to_unit_and_exps
            )
        )
        if self._unknown_unit_caption:
            lst.append(self._unknown_unit_caption)
        else:
            lst.append(None)
        return _ObtainReduced, (lst,)

    def __init__(self, category, unit, unknown_unit_caption=None):
        """
        :type category: unicode or odict
        :param category:
            The category to which the new quantity should be bound or an odict with information on
            the derived category/unit (in which case the unit parameter is ignored).

        :param unicode unit:
            The unit which the new quantity should have.
        """
        try:
            # Bail out if unit is already configured: this may happen when the Quantity constructor
            # is called directly: if Quantity() is called, the protocol will call __new__ and even
            # if a configured instance is returned, it'll call __init__ again, so, this is a check
            # for this situation so that we return if the quantity is already configured.
            self._unknown_unit_caption
            return
        except AttributeError:
            pass

        # Keep self._unit_database for faster access.
        unit_database = self._unit_database = UnitDatabase.GetSingleton()

        if unknown_unit_caption is not None:
            assert (
                unknown_unit_caption.__class__ == six.text_type
            ), "Unit caption must be a string. Note: the unit database and unit system parameters were removed."
            self._unknown_unit_caption = unknown_unit_caption
        else:
            self._unknown_unit_caption = ""

        if category.__class__ == odict:
            assert unit is None
            self._category_to_unit_and_exps = category
            self._is_derived = True

            rep_and_exp = odict()
            for category, (_unit, exp) in six.iteritems(
                self._category_to_unit_and_exps
            ):
                quantity_type = unit_database.GetCategoryQuantityType(category)
                existing = rep_and_exp.get(quantity_type, 0)
                rep_and_exp[quantity_type] = existing + exp

            self._category = self._MakeStr(
                [
                    (category, exp)
                    for category, (_unit, exp) in six.iteritems(
                        self._category_to_unit_and_exps
                    )
                ]
            )
            self._quantity_type = self._MakeStr(list(rep_and_exp.items()))
            self._unit = self._CreateUnitsWithJoinedExponentsString()
            self._composing_units = tuple(
                (unit, exp)
                for _category, (unit, exp) in six.iteritems(
                    self._category_to_unit_and_exps
                )
            )
            self._composing_categories = tuple(
                six.iterkeys(self._category_to_unit_and_exps)
            )
            self._category_info = None
            return

        self._is_derived = False

        # changes to accommodate making operations with derived units (internally the
        # information must be a dict)
        if category.__class__ != six.text_type:
            raise TypeError("Only unicode is accepted. %s is not." % category.__class__)

        # Getting the category should be enough to check that it's valid.
        category_info = unit_database.GetCategoryInfo(category)
        self._category_info = category_info

        # if no unit was given, assume base unit for this quantity_type
        if unit is None:
            unit = category_info.default_unit
        else:
            if unit.__class__ != six.text_type:
                raise TypeError("Only unicode is accepted. %s is not." % unit.__class__)
            unit_database.CheckCategoryUnit(category, unit)

        # store it as odict so that we can have a decent order when creating a string from
        # an operation.
        category_to_unit_and_exps = odict()
        category_to_unit_and_exps[category] = [unit, 1]
        self._category_to_unit_and_exps = category_to_unit_and_exps

        self._category = category
        self._quantity_type = unit_database.GetCategoryQuantityType(category)
        self._unit = unit
        self._tobase = self._unit_database.GetInfo(
            self._quantity_type, self._unit, fix_unknown=True
        ).tobase
        self._composing_units = unit
        self._composing_categories = category

    def MakeCopy(self, resulting_class=None, category_to_unit_and_exps=None):
        """
        Creates a new copy of this instance

        :param type resulting_class:
            Deprecated (no longer used).

        .. see:: _CreateDerived for parameter.

        :rtype: cls or resulting_class
        :returns:
            An copy of the current instance.
        """
        if category_to_unit_and_exps is None:
            return self  # As it's immutable, we can just return itself!

        # when creating it, a deepcopy of _category_to_unit_and_exps is done
        ret = Quantity._CreateDerived(
            category_to_unit_and_exps,
            resulting_class=resulting_class,
            validate_category_and_units=False,
            unknown_unit_caption=self._unknown_unit_caption,
        )
        return ret

    def CreateCopyInstance(self, category_to_unit_and_exps=None):
        """
        Creates a copy of this quantity with a new category/unit.

        .. see:: _CreateDerived for parameter.
        """
        if category_to_unit_and_exps is None:
            return self  # As it's immutable, we can just return itself!
        return self.MakeCopy(None, category_to_unit_and_exps)

    def Copy(self):
        """
        Create a copy of this class.
        """
        return self  # As it's immutable, we can just return itself!

    def __repr__(self, *args, **kwargs):
        if self._unknown_unit_caption:
            return "Quantity(%r, %r, %r)" % (
                self.GetCategory(),
                self._unit,
                self._unknown_unit_caption,
            )

        return "Quantity(%r, %r)" % (self.GetCategory(), self._unit)

    def SetUnknownCaption(self, caption):
        raise ReadOnlyError("Quantity is now read-only.")

    def GetUnknownCaption(self):
        return self._unknown_unit_caption

    def GetUnitCaption(self):
        unit = self._unit
        if unit == UNKNOWN_UNIT and self._unknown_unit_caption:
            # For now let's leave both (may need further discussions)
            unit = self._unknown_unit_caption + " " + UNKNOWN_UNIT

        return unit

    def GetCategory(self):
        return self._category

    def _MakeStr(self, repr_and_exp):
        """
        Used to make a string representation given a string and its exponents

        :type repr_and_exp: list(tuple(unicode, int))
        :param repr_and_exp:
            List of string, exponent.

        :rtype: unicode
        :returns:
            A string with the string and the given exponents. E.g.:
            Receiving [('m', 2)] will return m ** 2
            Receiving [('cm', -1)] will return 1 / cm
            Exponents = 0 are not added to the returned string.
        """
        ret = ""

        for rep, exp in repr_and_exp:
            if exp > 0:
                if ret:
                    ret += " * "

                if exp != 1:
                    ret += "(%s) ** %s" % (rep, exp)
                else:
                    ret += rep

        added_div = False
        for rep, exp in repr_and_exp:
            if exp < 0:
                if not added_div:
                    added_div = True
                    if ret:
                        ret += " / "
                    else:
                        ret += "1 / "

                if exp != -1:
                    ret += "(%s) ** %s" % (rep, abs(exp))
                else:
                    ret += rep

        return ret

    def GetQuantityType(self):
        return self._quantity_type

    def _CreateUnitsWithJoinedExponentsString(self):
        """
        Create a string with the joined units and exponents of the units to be shown to the
        user.

        :rtype: unicode
        :returns:
            A string with the units with the joined exponents.
            E.g.: m2, 1/m and so on (dependent on the unit it has internally)
        """
        ret = ""
        composing_units = self.GetComposingUnitsJoiningExponents()
        for unit, exp in composing_units:
            if exp > 0:
                if ret:
                    ret += "."

                if exp != 1:
                    ret += "%s%s" % (unit, exp)
                else:
                    ret += unit

        added_div = False
        for unit, exp in composing_units:
            if exp < 0:
                if not added_div:
                    added_div = True
                    if ret:
                        ret += "/"
                    else:
                        ret += "1/"

                ret += unit
                if exp != -1:
                    ret += six.text_type(abs(exp))

        return ret

    def GetUnit(self):
        return self._unit

    unit = property(GetUnit)

    def GetUnitName(self):
        """
        :rtype: unicode
        :returns:
            A description of the unit showing all the containing parts in the category,
            units and the exponent for each. This differs from the default unit because the
            default GetUnit will only return the final name of the unit, composing all the
            available parts into a single exponent (e.g. if it's composed of more than one
            'm', each with exponent 1, it would appear 2 times in this method and would appear
            'm2' in the GetUnit).
        """
        repr_and_exp = odict()
        unit_database = self._unit_database

        for category, (unit, exp) in six.iteritems(self._category_to_unit_and_exps):
            quantity_type = unit_database.GetCategoryQuantityType(category)
            unit_name = unit_database.GetUnitName(quantity_type, unit)
            existing = repr_and_exp.get(unit_name, 0)
            repr_and_exp[unit_name] = existing + exp

        return self._MakeStr(list(repr_and_exp.items()))

    def GetValidUnits(self):
        """
        Shortcut for getting the valid units

        :rtype: list(unicode)
        :returns:
            The valid units.
        """
        return self.GetUnitDatabase().GetValidUnits(self.GetCategory())

    def IsDerived(self):
        """
        :rtype: bool
        :returns:
            Returns whether this quantity has derived units.
        """
        return self._is_derived

    def GetCategoryInfo(self):
        """
        :rtype: CategoryInfo
        :returns:
            Returns the category info associated with this quantity.
        """
        return self._category_info

    def ConvertScalarValue(self, value, to_unit):
        """
        Convert a float to another unit (note: Convert may be used to convert non-float values).

        :param float value:
            The value to be converted.

        :type to_unit: unicode or list((unicode, int))
        :param to_unit:
            The target unit for the value.

        :rtype: float
        :returns:
            Returns the scalar converted to the passed unit.
        """
        if not self._is_derived:
            from_unit = self._unit

            # same unit: no conversion needed
            if from_unit == to_unit:
                return value

            other = self._unit_database.GetInfo(
                self._quantity_type, to_unit, fix_unknown=True
            )

            return other.frombase(self._tobase(value))
        else:
            return self.Convert(value, to_unit)

    def Convert(self, value, to_unit):
        """
        Convert any value object which has a conversion registered (note: ConvertScalarValue may
        be used to convert a float value in a more optimized way).

        :param object value:
            The value to be converted (array, numpy, etc.)

        :type to_unit: unicode or list((unicode, int))
        :param to_unit:
            The target unit for the value.

        :rtype: object
        :returns:
            An object with values to the passed unit.
        """
        return self._unit_database.Convert(
            self._composing_categories, self._composing_units, to_unit, value
        )

    @classmethod
    def _GetComparison(cls, operator, use_literals=False):
        """
        Method for getting different representations of comparisons for messages, using operator or
        literals.

        :param unicode operator:
            The key to the operator.

        :param Boolean use_literals:
            If literals are to be used.

        :returns unicode:
            A comparison representation, with operator or literal.
        """
        OPERATOR_COMPARISON = {">": ">", "<": "<", ">=": ">=", "<=": "<="}
        LITERALS_COMPARISON = {
            ">": "greater than",
            "<": "less than",
            ">=": "greater or equal to",
            "<=": "less or equal to",
        }
        if use_literals:
            return LITERALS_COMPARISON[operator]

        return OPERATOR_COMPARISON[operator]

    def _RaiseValueError(self, value, operator, limit_value, use_literals):
        """
        Raises a ValueError exception with a formatted message about the comparison of the value
        with it's limit.

        :param float value:
            Value with boundary error.

        :param unicode operator:
            The key to the operator.

        :param float limit_value:
            Value that defines the limit of the category.

        :param Boolean use_literals:
            If literals are to be used in the error message.
        """
        invalid_value_message = "Invalid value for %s: %g. Must be %s %r."
        raise ValueError(
            invalid_value_message
            % (
                self._category_info.caption,
                value,
                self._GetComparison(operator, use_literals),
                limit_value,
            )
        )

    def CheckValue(self, value, use_literals=False):
        """
        Checks if the passed value is valid for this quantity.

        :param float value:
            The value to be checked.

        :raises ValueError:
            if the value is not valid for this quantity.
        """
        if self._is_derived:  # It's not possible to check value for a derived category.
            return

        category_info = self._category_info
        unit = self._unit

        # checking minimum value
        if category_info.min_value is not None or category_info.max_value is not None:
            # convert value to check limit
            if unit != category_info.default_unit:
                value = self.ConvertScalarValue(value, category_info.default_unit)

            if category_info.min_value is not None:
                if category_info.is_min_exclusive:
                    if not value > category_info.min_value:
                        self._RaiseValueError(
                            value, ">", category_info.min_value, use_literals
                        )
                else:
                    if not value >= category_info.min_value:
                        self._RaiseValueError(
                            value, ">=", category_info.min_value, use_literals
                        )

            # checking maximum value
            if category_info.max_value is not None:
                if category_info.is_max_exclusive:
                    if not value < category_info.max_value:
                        self._RaiseValueError(
                            value, "<", category_info.max_value, use_literals
                        )
                else:
                    if not value <= category_info.max_value:
                        self._RaiseValueError(
                            value, "<=", category_info.max_value, use_literals
                        )

    def GetUnitDatabase(self):
        return self._unit_database

    def GetCategoryToUnitAndExps(self):
        return self._category_to_unit_and_exps

    def GetComposingCategories(self):
        return self._composing_categories

    def GetComposingUnits(self):
        return self._composing_units

    def GetComposingUnitsJoiningExponents(self):
        try:
            return self._composing_units_joining_exponents
        except AttributeError:
            ret = odict()
            for _category, (unit, exp) in six.iteritems(
                self._category_to_unit_and_exps
            ):
                existing = ret.get(unit, 0)
                ret[unit] = existing + exp
            self._composing_units_joining_exponents = tuple(ret.items())
        return self._composing_units_joining_exponents

    def GetCategoryToUnitAndExpsCopy(self, unit_and_exps=None):
        """
        Same as Quantity.GetCategoryToUnitAndExps but returns a copy instead of the internal
        instance.

        .. see:: Quantity.GetCategoryToUnitAndExps
        """
        if unit_and_exps is None:
            unit_and_exps = self._category_to_unit_and_exps

        return odict(
            (category, unit_and_exp[:])
            for (category, unit_and_exp) in six.iteritems(unit_and_exps)
        )

    def __hash__(self):
        try:
            return self._hash
        except AttributeError:
            lst = list(
                (category, tuple(unit_and_exp))
                for (category, unit_and_exp) in six.iteritems(
                    self._category_to_unit_and_exps
                )
            )
            lst.append(self._unknown_unit_caption)
            self._hash = hash(tuple(lst))
        return self._hash

    def __eq__(self, other):
        """
        Compares if this instance is equivalent to another quantity

        :param object other:
            The object we want to compare to

        :rtype: boo
        :returns:
            True if they're equal and False otherwise
        """
        if not other.__class__ == _Quantity:
            return False

        return (
            tuple(self._category_to_unit_and_exps.items())
            == tuple(other._category_to_unit_and_exps.items())
            and self._unknown_unit_caption == other._unknown_unit_caption
        )

    def __ne__(self, other):
        """
        The negation on __eq__

        .. see:: __eq__
        """
        return not self == other

    # Used to cast a constant value
    BASE_NUMBER = six.integer_types + (float,)

    OPERATION_DIVIDE = "Divide"
    OPERATION_MULTIPLY = "Multiply"
    OPERATION_SUBTRACT = "Subtract"
    OPERATION_ADD = "Sum"

    # lonly ----------------------------------------------------------------------------------------
    def __abs__(self):
        return self

    # right ----------------------------------------------------------------------------------------

    def __rmod__(self, other):
        return self.__truediv__(other)

    def __rmul__(self, other):
        return self._DoOperation(other, self, self.OPERATION_MULTIPLY)

    def __rsub__(self, other):
        return self._DoOperation(other, self, self.OPERATION_SUBTRACT)

    def __radd__(self, other):
        return self._DoOperation(other, self, self.OPERATION_ADD)

    # basic ----------------------------------------------------------------------------------------
    def __truediv__(self, other):
        return self._DoOperation(self, other, self.OPERATION_DIVIDE)

    if six.PY2:
        __div__ = __truediv__  # python2 uses __div__ as the / operator
    __mod__ = __truediv__
    __rtruediv__ = __truediv__

    def __mul__(self, other):
        return self._DoOperation(self, other, self.OPERATION_MULTIPLY)

    def __sub__(self, other):
        return self._DoOperation(self, other, self.OPERATION_SUBTRACT)

    def __add__(self, other):
        return self._DoOperation(self, other, self.OPERATION_ADD)

    def __pow__(self, exponent):
        result = self
        for _ in range(exponent - 1):
            result = self._DoOperation(self, result, self.OPERATION_MULTIPLY)
        return result

    def _DoOperation(self, q1, q2, operation):
        """
        Performs the given operation on the quantities returning the result

        :param Quantity q1:
            The left side operation quantity

        :param Quantity q2:
            The right side operation quantity

        :param OPERATION_* operation:
            The operation to be performed

        :rtype: Quantity
        :returns:
            The resulting quantity from the operation with the proper type and unit
            e.g. ('length', 'm') * ('length', 'm') = ('length **2', 'm2')
        """
        if isinstance(q1, self.BASE_NUMBER):
            return q2
        elif isinstance(q2, self.BASE_NUMBER):
            return q1
        elif isinstance(q1, Quantity) and isinstance(q2, Quantity):
            unit_database = self.GetUnitDatabase()
            operation = getattr(unit_database, operation)
            q, _v = operation(q1, q2, 1.0, 1.0)
            return q
        else:
            return NotImplemented



"""
    REMARKS: Unicode
        The "unit" must not containt unicode character. All unit name and representation must use
        the standard ascii representation.
"""
from __future__ import absolute_import, division, unicode_literals

import pytest
from pytest import approx

import six

from barril._foundation.odict import odict
from barril import units
from barril.units import (
    InvalidOperationError,
    InvalidUnitError,
    ObtainQuantity,
    Quantity,
    Scalar,
)


def testScalarInterface(unit_database_well_length):
    s = Scalar("well-length")
    assert 0 == s.value
    assert "m" == s.unit


def testQuantity(unit_database_well_length):
    Quantity("well-length", "m")
    with pytest.raises(units.InvalidQuantityTypeError):
        Quantity("foo", "m")


def testQuantityEq(unit_database_well_length):
    q0 = ObtainQuantity("m", "well-length")
    q1 = ObtainQuantity("m", "well-length")
    assert q0 == q1


def testReadOnlyQuantityCopy(unit_database_well_length):
    q0 = ObtainQuantity("m", "well-length")
    q1 = q0.Copy()
    assert q0 == q1


def testScalar(unit_database_well_length):
    q = ObtainQuantity("m", "well-length")
    s = Scalar.CreateWithQuantity(q, 10)

    assert 10 == s.value
    assert approx(abs(393.700787402 - s.GetValue(unit="in")), 7) == 0

    s = s.CreateCopy(value=0.254, unit="in")
    assert 0.254 == s.value
    assert "in" == s.unit
    assert "well-length" == s.category

    with pytest.raises(AttributeError):
        setattr(s, "value", 10)


def GetQuantity(unit="cm"):
    return units.ObtainQuantity(unit, "well-length")


def testBadUnit(unit_database_well_length):
    x = units.Scalar.CreateWithQuantity(GetQuantity(), 500)
    with pytest.raises(units.InvalidUnitError):
        x.CreateCopy(unit="invalidUnit")


def testOperations(unit_database_well_length):
    x = units.Scalar.CreateWithQuantity(GetQuantity(), 500)
    assert x.value == 500
    assert x.unit == "cm"

    assert x.GetValue("m") == 5
    assert x.unit == "cm"
    assert x.GetValue("cm") == 500
    assert x.unit == "cm"
    assert x.GetValue("km") == 0.005


def unit(value, unit="m"):
    return units.Scalar.CreateWithQuantity(GetQuantity(unit), value)


def testEq(unit_database_well_length):
    u1 = units.Scalar.CreateWithQuantity(GetQuantity(), 1.0)
    u2 = units.Scalar.CreateWithQuantity(GetQuantity(), 1.0)
    assert u1 == u2
    assert u1.unit == u2.unit
    assert "cm" == u2.unit


def testAddUnit(unit_database_empty):
    unit_database = unit_database_empty
    unit_database.AddUnitBase("length", "metre", "m")
    RATIO = 0.0254
    unit_database.AddUnit(
        "length",
        "inch",
        "in",
        lambda x: 0.0254 * RATIO,
        lambda y: y / RATIO,
        default_category=None,
    )
    unit_database.AddCategory(
        "length", "length", override=True, valid_units=["m", "in"]
    )
    unit_database.AddCategory("depth", "length", override=True, valid_units=["m", "in"])

    a = units.Scalar("length", value=1.0, unit="m")
    b = units.Scalar("length", value=2.0, unit="m")
    c = units.Scalar("length", value=1.0, unit="in")
    d = units.Scalar("depth", value=1.0, unit="m")
    e = None
    f = units.Scalar("length", value=1.0, unit="m")

    assert a != b
    assert a != c
    assert a != e
    assert a != d

    assert a == f


def testValidUnits(unit_database_empty):
    # let'scalar clear up the unit manager
    unit_database = unit_database_empty

    unit_database.AddUnitBase("length", "meters", "m")
    unit_database.AddUnit("length", "centimeters", "cm", "%f * 100.0", "%f / 100.0")
    unit_database.AddUnit("length", "milimeters", "mm", "%f * 1000.0", "%f / 1000.0")
    unit_database.AddUnit("length", "kilometers", "km", "%f / 1000.0", "%f * 1000.0")
    unit_database.AddUnit("length", "miles", "mi", "%f / 1609.347", "%f * 1609.347")
    unit_database.AddUnit("length", "inches", "in", "%f / 0.0254", "%f * 0.0254")

    valid_units = ["m", "cm", "mm"]
    unit_database.AddCategory("length", "length")
    unit_database.AddCategory("well-length", "length", valid_units=valid_units)
    unit_database.AddCategory("well-diameter", "length")

    Quantity("well-length", "km")

    scalar = units.Scalar("well-length", 1, "m")
    assert valid_units == scalar.GetValidUnits()

    scalar = units.Scalar("well-diameter", 1, "m")
    assert ["m", "cm", "mm", "km", "mi", "in"] == scalar.GetValidUnits()

    # Creating a scalar in a unit that isn't valid shouldn't raise an error and the unit
    # will be added to valid units
    scalar = units.Scalar("well-length", 1, "mi")
    assert scalar.GetValidUnits() == ["m", "cm", "mm", "mi"]


def testScalarCopyAndRepresentation(unit_database_empty):
    unit_database = unit_database_empty

    unit_database.AddUnitBase("length", "meters", "m")
    unit_database.AddUnit("length", "milimeters", "mm", "x * 1000.0", "x / 1000.0")
    unit_database.AddUnit("length", "centimeters", "cm", "x * 100.0", "x / 100.0")
    unit_database.AddCategory("well-diameter", "length")
    s = Scalar("well-diameter", 10, "m")

    assert "Scalar(10.0, 'm', 'well-diameter')" == repr(s)
    assert "10.0" == six.text_type(s.value)
    assert "m" == s.unit
    assert "well-diameter" == s.GetCategory()
    assert "length" == s.GetQuantityType()

    s = s.CreateCopy(unit="cm")
    assert "Scalar(1000.0, 'cm', 'well-diameter')" == repr(s)
    assert "1000 [cm]" == six.text_type(s)


def testSort(unit_database_empty):
    db = unit_database_empty

    db.AddUnitBase("length", "meters", "m")
    db.AddUnit("length", "centimeters", "cm", "x * 100.0", "x / 100.0")
    db.AddCategory("length", "length")

    s1 = Scalar("length", 10.0, "cm")
    s2 = Scalar("length", 1.0, "m")
    s3 = Scalar("length", 500.0, "cm")
    s4 = Scalar("length", 10.0, "m")

    x = [s2, s1, s4, s3]
    x.sort()
    assert x == [s1, s2, s3, s4]


def testCategoryParameters(unit_database_empty):
    db = unit_database_empty

    db.AddUnitBase("length", "meters", "m")
    db.AddUnit("length", "centimeters", "cm", "x * 100.0", "x / 100.0")

    db.AddCategory(
        "my length",
        "length",
        ["cm", "m"],
        default_unit="m",
        default_value=18.5,
        min_value=-30.0,
        max_value=100.0,
        is_min_exclusive=True,
        is_max_exclusive=False,
    )

    s1 = Scalar("my length")
    assert s1.value == 18.5
    assert s1.unit == "m"

    # test maximum boundary
    s2 = Scalar("my length", 100.0, "m")
    assert s2.value == 100.0
    assert s2.unit == "m"

    s3 = Scalar("my length", 500.0, "cm")
    assert s3.value == 500.0
    assert s3.unit == "cm"

    # higher than maximum value
    another = Scalar("my length", 120.0, "m")
    assert not another.IsValid()
    with pytest.raises(ValueError):
        another.CheckValidity()

    another = Scalar("my length", -30.0, "m")
    assert not another.IsValid()

    s4 = Scalar("my length", unit="cm")
    assert s4.value == 1850.0
    assert s4.unit == "cm"

    another = s4.CreateCopy(12500.00, "cm")
    assert not another.IsValid()


def testInvalidUnit(unit_database_empty):
    unit_database = unit_database_empty

    unit_database.AddUnitBase("length", "meters", "m")
    unit_database.AddUnit("length", "milimeters", "mm", "x * 1000.0", "x / 1000.0")
    unit_database.AddUnit("length", "centimeters", "cm", "x * 100.0", "x / 100.0")
    unit_database.AddCategory("well-diameter", "length")
    with pytest.raises(InvalidUnitError):
        Scalar("well-diameter", 10, "days")


def testDerivedUnits(unit_database_empty):
    unit_database = unit_database_empty

    unit_database.AddUnitBase("length", "meters", "m")
    unit_database.AddUnit("length", "milimeters", "mm", "x * 1000.0", "x / 1000.0")
    unit_database.AddUnit("length", "centimeters", "cm", "x * 100.0", "x / 100.0")
    unit_database.AddCategory("well-diameter", "length")
    unit_database.AddUnitBase("time", "seconds", "s")
    unit_database.AddCategory(category="Time", quantity_type="time")

    s1 = Scalar("well-diameter", 10, "m")
    s2 = Scalar("well-diameter", 10, "cm")
    s3 = Scalar("Time", 10, "s")

    assert Scalar("well-diameter", 10.10, "m") == s1 + s2
    assert Scalar("well-diameter", 9.90, "m") == s1 - s2

    quantity = Quantity.CreateDerived(odict())
    assert Scalar.CreateWithQuantity(quantity, 100) == s1 / s2
    assert Scalar("well-diameter", 9.90, "m") == s1 - s2
    with pytest.raises(InvalidOperationError):
        s1.__sub__(s3)


def testCreationWithDerivedQuantity(unit_database_len_time):
    unit_database = unit_database_len_time

    m = Quantity.CreateDerived(odict([("Table size", ["m", 1])]))
    km_city = Quantity.CreateDerived(odict([("City size", ["km", 1])]))

    quantity, value = unit_database.Multiply(m, km_city, 1, 0.01)
    calculated1 = Scalar.CreateWithQuantity(quantity, value)
    assert six.text_type(calculated1)

    s1 = Scalar.CreateWithQuantity(m, 1)
    s2 = Scalar.CreateWithQuantity(km_city, 0.01)
    assert calculated1 == s1 * s2


def testNumberInteractions(unit_database_len_time):
    scalar = Scalar("Table size", 1, "m")
    scalar2 = Scalar.CreateWithQuantity(Quantity.CreateDerived(odict()), 0)

    assert scalar == scalar + scalar2

    assert scalar == scalar + 0
    assert scalar == 0 + scalar

    assert 9 == (10 - scalar).value
    assert -9 == (scalar - 10).value

    assert 10 == (10 * scalar).value
    assert 10 == (scalar * 10).value

    assert 10 == (10.0 / scalar).value
    assert 1 / 10.0 == (scalar / 10.0).value


def testDivision(unit_database_len_time):
    unit_database = unit_database_len_time

    m = Quantity.CreateDerived(odict([("Table size", ["m", 1])]))
    km_city = Quantity.CreateDerived(odict([("City size", ["km", 1])]))
    quantity, value = unit_database.Divide(m, km_city, 1, 0.01)
    calculated1 = Scalar.CreateWithQuantity(quantity, value)
    s1 = Scalar.CreateWithQuantity(m, 1)
    s2 = Scalar.CreateWithQuantity(km_city, 0.01)
    assert calculated1 == s1 / s2


def testPow(unit_database_len_time):
    s = Scalar(2, "s", "Time")
    spow = s ** 3
    assert spow.unit == "s3"
    assert spow.value == 8


def testFormatedUnitsOnScalar(unit_database_empty):
    """
        Allow present units on get formatted Scalar
    """
    unit_database = unit_database_empty
    unit_database.AddUnitBase("length", "meters", "m")
    unit_database.AddCategory(category="length", quantity_type="length")

    class MyScalar(units.Scalar):
        pass

    scalar = MyScalar("length", 1.18, "m")
    assert scalar.GetFormatted() == "1.18 [m]"

    MyScalar.SetFormattedValueFormat("%0.4f")
    MyScalar.SetFormattedSuffixFormat(" (%s)")

    # Tests the GET methods
    assert MyScalar.GetFormattedValueFormat() == "%0.4f"
    assert MyScalar.GetFormattedSuffixFormat() == " (%s)"

    # Tests the Formatted method
    assert scalar.GetFormatted() == "1.1800 (m)"

    # Check for invalid format strings
    with pytest.raises(TypeError):
        MyScalar.SetFormattedValueFormat("%d [%s]")
    with pytest.raises(TypeError):
        MyScalar.SetFormattedSuffixFormat("%s %s")
    with pytest.raises(TypeError):
        MyScalar.SetFormattedSuffixFormat("%f")

    # When invalid format are set, the formats are not changed
    assert MyScalar.GetFormattedValueFormat() == "%0.4f"
    assert MyScalar.GetFormattedSuffixFormat() == " (%s)"

    # Make sure we didn't affect the Scalar class, only the local MyScalar
    MyScalar.SetFormattedValueFormat("<<< %f")
    MyScalar.SetFormattedSuffixFormat(" [%s] >>>")
    scalar = Scalar("length", 1.18, "m")
    assert scalar.GetFormatted() == "1.18 [m]"


def testEmptyScalar():
    """
        ScalarMultiData exception when some of its scalars don't have a quantity_type
    """
    # An empty scalar doesn't have a category defined
    scalar_1 = Scalar.CreateEmptyScalar(20.0)

    # When try to retrieve scalar value or unit a  exception was being raised
    assert scalar_1.GetValue("m") == 20.
    assert scalar_1.GetUnit() == ""


def testCopyProperties(unit_database_well_length):
    """
        Test if the mehod SetValue is not called when copying the Scalar's properties.
    """
    category = "well-length"
    unit = "m"
    value = 10.0

    scalar_source = Scalar(category, value, unit)
    scalar_dest = scalar_source.CreateCopyInstance()

    assert scalar_dest.GetCategory() == category
    assert scalar_dest.GetUnit() == unit
    assert scalar_dest.GetValue() == value


def testCopyPropertiesAndValidation(unit_database_well_length):
    category = "well-length-with-min-and-max"  # the minimum value is zero
    unit = "m"
    value = -1

    # Not raises exception because validation is False on the copy operation
    scalar_source = Scalar(category, value, unit)
    assert not scalar_source.IsValid()

    # Raises exception because validation is True and the value -1 is invalid accordingly to
    # the category 'well-length-with-min-and-max'
    another = scalar_source.CreateCopy()
    assert not another.IsValid()


def testDefaultValue(unit_database_len_pressure):
    """
        Scalar constructor considers the minimum and maximum values
        when default_value is not defined
    """
    db = unit_database_len_pressure

    db.AddCategory(
        category="my length", quantity_type="length", min_value=100.0, max_value=200.0
    )

    # if the default value is not defined, the scalar should not try to set the initial/first
    # value of the new instance to 0.0, but it should assume the minimum allowed value.
    # should not raise ValueError
    length = Scalar("my length")
    assert length.GetValue() == 100.0

    length = Scalar(ObtainQuantity("m", "my length"))
    assert length.GetValue() == 100.0

    length = Scalar(ObtainQuantity("m"))
    assert length.GetValue() == 0.0

    # invalid default value (< min)
    with pytest.raises(AssertionError):
        db.AddCategory(
            category="my pressure",
            quantity_type="pressure",
            default_value=50.0,
            min_value=100.0,
            max_value=200.0,
        )

    # invalid default value (> max)
    with pytest.raises(AssertionError):
        db.AddCategory(
            category="my pressure",
            quantity_type="pressure",
            default_value=300.0,
            min_value=100.0,
            max_value=200.0,
        )

    # invalid default value (<= min)
    with pytest.raises(AssertionError):
        db.AddCategory(
            category="my pressure",
            quantity_type="pressure",
            default_value=100.0,
            min_value=100.0,
            max_value=200.0,
            is_min_exclusive=True,
        )

    # invalid default value (>= min)
    with pytest.raises(AssertionError):
        db.AddCategory(
            category="my pressure",
            quantity_type="pressure",
            default_value=200.0,
            min_value=100.0,
            max_value=200.0,
            is_max_exclusive=True,
        )

    db.AddCategory(
        category="my pressure",
        quantity_type="pressure",
        min_value=100.0,
        max_value=200.0,
        default_value=150.0,
        is_min_exclusive=True,
        is_max_exclusive=False,
    )

    pressure = Scalar("my pressure")
    assert pressure.GetValue() == 150.0

    # default_value not informed. checking if the interval limits are respected
    # when is_min_exclusive or is_max_exclusive have been informed, so the default_value
    # must be
    with pytest.raises(RuntimeError):
        db.AddCategory(
            category="my pressure 2",
            quantity_type="pressure",
            min_value=100.0,
            max_value=200.0,
            is_max_exclusive=True,
        )


def testScalarInvalidValue(unit_database_len_time):
    db = unit_database_len_time

    db.AddCategory("another-length", "length", min_value=0, max_value=15)

    scalar = Scalar("another-length", value=15, unit="m")
    assert scalar.IsValid()

    scalar = Scalar("another-length", value=-5, unit="m")
    assert not scalar.IsValid()

    # By default the validation will be performed. 15 is a valid value.
    scalar = Scalar("another-length", value=15, unit="m")
    assert scalar.IsValid()

    # Even invalid ,the scalar returns the value, unit and a formatted text.
    another = Scalar("another-length", value=3000, unit="m")
    assert not another.IsValid()
    assert another.GetValue("m") == 3000
    assert another.GetUnit() == "m"
    assert another.GetFormatted() == "3000 [m]"

    # By default the validation will be performed, and in this cases will raise ValueError.
    another_2 = Scalar("another-length", unit="m", value=5000)
    assert not another_2.IsValid()

    # Performing copy between invalid scalars. The validation is not performed on copy.
    copied = another.CreateCopy(unit="cm")
    assert not copied.IsValid()
    assert copied.GetValue("m") == 3000
    assert copied.GetUnit() == "cm"
    assert copied.GetFormatted() == "300000 [cm]"


def testScalarValidUnits(unit_database_empty):
    db = unit_database_empty
    db.AddUnit("length", "milimeters", "mm", "%f * 1000.0", "%f / 1000.0")
    db.AddUnitBase("length", "meters", "m")
    db.AddUnit("length", "centimeters", "cm", "%f * 100.0", "%f / 100.0")
    db.AddUnit("length", "kilometers", "km", "%f / 1000.0", "%f * 1000.0")
    db.AddCategory("len1", "length", valid_units=["mm", "cm"])
    db.AddCategory("len2", "length", valid_units=["m", "km"])

    s1 = Scalar("len1", 1, "mm")
    assert s1 is not None

    s2 = Scalar("len2", 1, "mm")
    assert s2 is not None


def testScalarRepresentation(unit_database_posc):
    s1 = Scalar(0.5, "kg/kg", "dimensionless")
    s2 = Scalar(0.125, "g/g", "mass concentration")

    assert s1.GetCategory() == "dimensionless"
    assert s2.GetCategory() == "mass concentration"

    assert repr(s1) == "Scalar(0.5, 'kg/kg', 'dimensionless')"
    assert repr(s2) == "Scalar(0.125, 'g/g', 'mass concentration')"


def testScalarCreationModes():
    base = Scalar(10, "m")

    assert Scalar("length", 10, "m") == base
    assert Scalar((10.0, "m")) == base

    with pytest.raises(AssertionError):
        Scalar("length", 1.0)  # missing unit


def testScalarPickle(unit_database_posc):

    import pickle

    simple_scalar = Scalar("length", 10, "m")
    simple_scalar2 = pickle.loads(pickle.dumps(simple_scalar))
    assert simple_scalar == simple_scalar2

    complex_scalar = Scalar("length", 10, "m") * Scalar("time", 5, "s")
    complex_scalar2 = pickle.loads(pickle.dumps(complex_scalar))
    assert complex_scalar == complex_scalar2


def testScalarHashEq():
    scalar1 = Scalar("length", 10, "m")
    scalar2 = Scalar("length", 10, "m")
    scalar3 = Scalar("length", 10, "cm")
    scalar4 = Scalar("length", scalar2.GetValue("cm"), "cm")

    assert scalar1 == scalar2
    assert hash(scalar1) == hash(scalar2)
    assert hash(scalar1) == hash(scalar1)
    assert scalar2 != scalar3
    assert hash(scalar2) != hash(scalar3)
    assert scalar3 != scalar4
    assert hash(scalar4) != hash(scalar3)



from __future__ import absolute_import, unicode_literals

import pickle

from barril._foundation.odict import odict
from barril.units import (
    UNKNOWN_QUANTITY,
    UNKNOWN_QUANTITY_TYPE,
    UNKNOWN_UNIT,
    GetUnknownQuantity,
    ObtainQuantity,
    Quantity,
)
from barril.units.unit_database import InvalidUnitError, UnitDatabase

import pytest


def testQuantityInit(mocker):

    # 1: cache it
    ObtainQuantity("m", "length")

    # 2: check
    calls = [0]
    original = UnitDatabase.GetSingleton

    def _New(*args, **kwargs):
        calls[0] += 1
        return original()

    mocker.patch.object(UnitDatabase, "GetSingleton")
    UnitDatabase.GetSingleton.return_value = _New()

    ObtainQuantity("m", "length")
    assert calls[0] == 1


def testQuantitySharedInstances():
    quantity = ObtainQuantity("m", "length")
    assert quantity is ObtainQuantity("m", "length")


def testQuantityCaption(unit_database_posc_len):
    unit_database = unit_database_posc_len
    unit_database.AddUnitBase(
        UNKNOWN_QUANTITY_TYPE, UNKNOWN_QUANTITY_TYPE, UNKNOWN_UNIT
    )
    unit_database.AddCategory(UNKNOWN_QUANTITY_TYPE, UNKNOWN_QUANTITY_TYPE)

    q0 = ObtainQuantity(unit=UNKNOWN_UNIT, category=UNKNOWN_QUANTITY_TYPE)
    assert "" == q0.GetUnknownCaption()

    q = ObtainQuantity(
        unit=UNKNOWN_UNIT, unknown_unit_caption="Feeeet", category=UNKNOWN_QUANTITY_TYPE
    )
    assert "Feeeet" == q.GetUnknownCaption()
    assert "Feeeet <unknown>" == q.GetUnitCaption()

    q = Quantity.CreateDerived(
        odict([("length", ("m", 1))]), unknown_unit_caption="Feeeet"
    )
    assert "m" == q.GetUnitCaption()


def testUnknownQuantity():
    global_unknown = UNKNOWN_QUANTITY

    # Global
    assert global_unknown is GetUnknownQuantity()

    # Cache
    my_unknown = GetUnknownQuantity("my_unknown")
    assert my_unknown is GetUnknownQuantity("my_unknown")


def testHash():
    quantity1 = ObtainQuantity("m")
    quantity2 = ObtainQuantity("m", "length")
    assert hash(quantity1) == hash(quantity2)
    assert quantity1 == quantity2

    unknown1 = ObtainQuantity(UNKNOWN_UNIT, unknown_unit_caption="foo")
    unknown2 = ObtainQuantity(UNKNOWN_UNIT, unknown_unit_caption="foo_bar")
    unknown3 = ObtainQuantity(UNKNOWN_UNIT, unknown_unit_caption="foo")
    assert hash(unknown1) == hash(unknown3)
    assert unknown1 == unknown3

    assert unknown1 != unknown2


def testPickle():
    quantity1 = ObtainQuantity("m")
    quantity2 = quantity1.__reduce__()
    assert quantity1 is quantity2[0](*quantity2[1])
    obtained = pickle.loads(pickle.dumps(quantity1))
    assert obtained is quantity1, "%s != %s" % (obtained, quantity1)


def testConvert():
    assert [1000.0, 3000.0] == ObtainQuantity("m").Convert([10, 30], "cm")


def testUnknown():
    with pytest.raises(InvalidUnitError):
        ObtainQuantity("m3", UNKNOWN_QUANTITY_TYPE)


def testDivision():
    meters = ObtainQuantity("m")
    seconds = ObtainQuantity("s")
    speed = meters / seconds
    no_unit = meters / meters
    assert speed.GetUnit() == "m/s"
    assert speed.unit == "m/s"
    assert no_unit.GetUnit() == ""



"""
B{Unit Management}

It is very common within an application (especially on engineering applications) to have
several different values within the application, regarding pressure, density, etc. But just
having the values without any scope is something that can be meaningless, as that value only
makes sense if we have a unit and we know what it relates to.

This module is designed to help the application manage those values and units, based on the
following B{requisites}:

  - changing all the units in the application that relate to something to a different unit
    (e.g.: all the units that relate to the well height that are currently in the application
    as meters should be treated with km from now on).
  - being able to easily change the unit on some entity and have its value updated (in python)
  - getting all the created elements that have some kind of measure, such as length or depth.
  - having the info on C++ is a requisite, but the automatic conversion utilities are not, as C++
    is used for hardcore processing, this is usually done manually (as a generic way would waste
    too much processing), so, basically, the structure in C++ is important, but things related to
    the application management are not.
  - have a default way of managing those values within the application, and giving default
    implementations for it (such as Scalars, Arrays)
  - being able to define the units that can be used for some category

Usage:

    >>> from barril import units
    >>> from barril.units import scalar
    >>>
    >>> unit_manager = units.UnitDatabase()
    >>> unit_manager.AddUnitBase('length', 'meters', 'm')
    >>> unit_manager.AddUnit('length', 'milimeters', 'mm', 'x * 1000.0', 'x / 1000.0')
    >>> unit_manager.AddUnit('length', 'centimeters', 'cm', 'x * 100.0', 'x / 100.0')
    >>> unit_manager.AddCategory('well-diameter', 'length')
    >>> s = scalar.Scalar.Create('well-diameter', 10, 'm', unit_manager)
    >>> repr(s)
    "'Scalar'('length', 10, 'm')"
    >>> print s.value
    10
    >>> print s.unit
    m
    >>> s.GetCategory()
    'well-diameter'
    >>> s.GetQuantityType()
    'length'
    >>> s.unit = 'cm'
    >>> repr(s)
    "'Scalar'('length', 1000.0, 'cm')"
    >>> print s
    1000.00 centimeters
    >>> print s.GetValue(unit='mm')
    10000.0


B{Design decisions}:

    - In order to make the validation of categories, quantity types and units, the unit-manager
    must be passed around when the objects are created and a weakref is kept for future changes.

    - It is not all 'static' as it was before because this would hinder some applications, especially
    applications with multiple documents, as those applications may need multiple unit-managers
    (one for each open document).


@see: L{definitions} for the basic interfaces defined.
"""

from __future__ import absolute_import, unicode_literals

from weakref import WeakValueDictionary

import six

from ._abstractvaluewithquantity import AbstractValueWithQuantityObject  # noqa
from ._array import Array  # noqa
from ._fixedarray import FixedArray  # noqa
from ._fraction_scalar import FractionScalar  # noqa
from ._quantity import ObtainQuantity, Quantity, ReadOnlyError  # noqa
from ._scalar import Scalar  # noqa
from ._scalar_factory import ScalarFactory  # noqa
from ._unit_constants import (  # noqa
    LENGTH_QUANTITY_TYPE,
    UNKNOWN_QUANTITY_TYPE,
    UNKNOWN_UNIT,
    CreateUnknwonwReadOnlyQuantity,
)
from .unit_database import (  # noqa
    InvalidOperationError,
    InvalidQuantityTypeError,
    InvalidUnitError,
    UnitDatabase,
    UnitInfo,
    UnitsError,
)

__all__ = [
    str("AbstractValueWithQuantityObject"),  # pylint: disable=invalid-all-object
    str("Array"),  # pylint: disable=invalid-all-object
    str("ChangeScalars"),  # pylint: disable=invalid-all-object
    str("FixedArray"),  # pylint: disable=invalid-all-object
    str("FractionScalar"),  # pylint: disable=invalid-all-object
    str("GetUnknownQuantity"),  # pylint: disable=invalid-all-object
    str("InvalidQuantityTypeError"),  # pylint: disable=invalid-all-object
    str("InvalidUnitError"),  # pylint: disable=invalid-all-object
    str("Quantity"),  # pylint: disable=invalid-all-object
    str("ReadOnlyError"),  # pylint: disable=invalid-all-object
    str("Scalar"),  # pylint: disable=invalid-all-object
    str("ScalarFactory"),  # pylint: disable=invalid-all-object
    str("UnitDatabase"),  # pylint: disable=invalid-all-object
    str("UnitsError"),  # pylint: disable=invalid-all-object
]

# Unknown quantity instance
UNKNOWN_QUANTITY = ObtainQuantity(UNKNOWN_UNIT, UNKNOWN_QUANTITY_TYPE)

# Unknown quantity weak value cache
UNKNOWN_QUANTITY_WEAK_CACHE = WeakValueDictionary()


# ===================================================================================================
# GetUnknownQuantity
# ===================================================================================================
def GetUnknownQuantity(unknown_caption=None):
    """
    Returns the quantity object for Unknown units.

    :rtype: Quantity
    """
    if unknown_caption:
        return ObtainQuantity(UNKNOWN_UNIT, UNKNOWN_QUANTITY_TYPE, unknown_caption)

    return UNKNOWN_QUANTITY


# ===================================================================================================
# Utilities
# ===================================================================================================
def ChangeScalars(owner, **scalars):
    """
    Change the given set of scalars for the owner

    :param owner: object
        The objectwith scalar instances to be changed.

    :param kwargs scalars:
        A dict with the scalar attribute names and the values and unit to be changed

    e.g.
        # To change value and unit
        ChangeScalars(fluid, density=(10, 'lbm/galUS'), 'concentration=(1.0, '%'))
        # To change value only
        ChangeScalars(fluid, density=(10, None))
        # To change unit only
        ChangeScalars(fluid, density=(None, 'lbm/galUS'))
    """
    for scalar_name, (value, unit) in six.iteritems(scalars):
        new_scalar = getattr(owner, scalar_name).CreateCopy(value=value, unit=unit)
        setattr(owner, scalar_name, new_scalar)



from __future__ import absolute_import, unicode_literals

import numpy
from six.moves import zip  # @UnresolvedImport


# ===================================================================================================
# _ValueGenerator
# ===================================================================================================
class _ValueGenerator(object):
    """
    Generator for values so that 1 one of the values is generated from iterating through a list and
    the other is a fixed values returned.

    Special case is if a numpy array is found... if it is, just yields the numpy array all at once
    (as it's able to make the whole range of operations without the need for an iteration in the
    python side).

    Internal API used to convert array values.
    """

    def __init__(self, p1, p2):
        self.p1 = p1
        self.p2 = p2

        self.iterate_1st = isinstance(p1, (tuple, list))
        self.iterate_2nd = isinstance(p2, (tuple, list))

    def IsNumpy(self):
        """
        :rtype: bool
        :returns:
            If either one of the contained elements is a numpy array.
        """
        return isinstance(self.p1, numpy.ndarray) or isinstance(self.p2, numpy.ndarray)

    def IsTuple(self):
        """
        :rtype: bool
        :returns:
            Whether the element(s) iterated is a tuple.
        """
        if self.iterate_1st and self.iterate_2nd:
            return isinstance(self.p1, tuple) and isinstance(self.p2, tuple)

        if self.iterate_1st:
            return isinstance(self.p1, tuple)

        if self.iterate_2nd:
            return isinstance(self.p2, tuple)

        return False

    def __iter__(self):
        """
        Yield the individual values we should convert (which may mean the same number all the time
        and the value being iterated in a list). Numpy is a special case (see classdocs)
        """
        if self.IsNumpy():
            yield self.p1, self.p2

        elif self.iterate_1st and not self.iterate_2nd:
            static_value = self.p2
            for v in self.p1:
                yield static_value, v

        elif not self.iterate_1st and self.iterate_2nd:
            static_value = self.p1
            for v in self.p2:
                yield v, static_value

        elif not self.iterate_1st and not self.iterate_2nd:
            yield self.p1, self.p2

        else:  # iterate both
            for v0, v1 in zip(self.p1, self.p2):
                yield v0, v1



from __future__ import absolute_import, unicode_literals

from barril.units.unit_rich_text_representations import UnitRichTextRepresentations


def AssertRTFCaption(received_representation, expected_caption):
    html_mask = UnitRichTextRepresentations.HTML_TEXT_MASK
    assert received_representation == html_mask % expected_caption


def testUnitRichTextRepresentations():

    # Some unit representation should be initially set
    expected_units_initially_set = [
        ("m3", "m<sup>3</sup>"),
        ("m3/d", "m<sup>3</sup>/d"),
        ("m3/m3", "m<sup>3</sup>/m<sup>3</sup>"),
    ]

    # Testing the default values
    for unit, expected_representation in expected_units_initially_set:
        stored_representation = UnitRichTextRepresentations.GetUnitHtmlRepresentation(
            unit
        )
        AssertRTFCaption(stored_representation, expected_representation)

    # If not found the representation in defaults, returns the unit
    unknown_unit_caption = "unknown_unit"
    unknown_unit_representation = UnitRichTextRepresentations.GetUnitHtmlRepresentation(
        unknown_unit_caption
    )
    assert unknown_unit_representation == unknown_unit_caption

    # It should be possible to set the rich text representation units
    new_unit = "new_unit"
    new_unit_representation = "new<sup>unit</sup>"

    UnitRichTextRepresentations.DEFAULT_UNITS_RICH_TEXT_REPRESENTATIONS[
        new_unit
    ] = new_unit_representation
    AssertRTFCaption(
        UnitRichTextRepresentations.GetUnitHtmlRepresentation(new_unit),
        new_unit_representation,
    )

    # But the user should be able to overwrite the default values with a custom representation
    new_cubic_meter_representation = "m<sub>3</sub>"
    cubic_meter_caption = "m3"
    UnitRichTextRepresentations.DEFAULT_UNITS_RICH_TEXT_REPRESENTATIONS[
        cubic_meter_caption
    ] = new_cubic_meter_representation

    AssertRTFCaption(
        UnitRichTextRepresentations.GetUnitHtmlRepresentation(cubic_meter_caption),
        new_cubic_meter_representation,
    )



from __future__ import absolute_import, unicode_literals

import copy
from functools import total_ordering

import six

from barril._foundation.types_ import CheckType
from barril.basic.fraction import FractionValue

from ._abstractvaluewithquantity import AbstractValueWithQuantityObject
from ._quantity import ObtainQuantity
from .unit_database import UnitDatabase


@total_ordering
class FractionScalar(AbstractValueWithQuantityObject):
    """
    Base class for objects similar to scalars, but that represent its value as L{FractionValue}
    instance instead of a float. Useful to describe diameters for
    pipes and wells in a more natural way to the user (for instance, "5 3/4 inches").
    """

    def _InternalCreateWithQuantity(self, quantity, value, unit_database=None):
        """
        For internal use only. Is used to initialize the actual quantity.

        :type quantity: unicode or IQuantity
        :param quantity:
            The quantity of this scalar.

        :param FractionValue value:
            The initial value
        """
        # Considering fraction values values are easily coerced from float values (though it is
        # important to note the opposite is not true) if the input value is not a fraction already
        # try to convert value to float. This also makes this subclass SetValue interface compatible
        # with superclass interface.
        try:
            if type(value) != FractionValue:
                value = FractionValue(number=float(value))
        except Exception:
            # If not a fraction and coercion to float fails, use CheckType to provide a better error
            # message.
            CheckType(value, (FractionValue, float))

        self._value = value
        self._quantity = quantity
        self._unit_database = unit_database or UnitDatabase.GetSingleton()

    def CheckValidity(self):
        """
        :raises ValueError: when current value is wrong somehow (out of limits, for example).
        """
        self._quantity.CheckValue(float(self._value))

    # Value ----------------------------------------------------------------------------------------
    def GetAbstractValue(self, unit=None):
        """

        :param unit:
        """
        if unit is None:
            return self._value

        return self.ConvertFractionValue(self._value, self._quantity, self.unit, unit)

    GetValue = GetAbstractValue
    value = property(GetAbstractValue)

    def _GetDefaultValue(self, category_info, unit=None):
        """

        :param category_info:
        :param unit:
        """
        value = category_info.default_value
        if unit is not None:
            # needs to convert value to default unit
            value = ObtainQuantity(
                category_info.default_unit, category_info.category
            ).ConvertScalarValue(value, unit)

        return value

    def GetValueAndUnit(self):
        return (self.GetValue(), self.GetUnit())

    @classmethod
    def ConvertFractionValue(cls, fraction_value, quantity, from_unit, to_unit):
        """
        Converts the given fraction value to the given unit.

        :type fraction_value: L{FractionValue}
        :param fraction_value:
            the fraction value to convert

        :type quantity: IQuantity or unicode
        :param quantity:
            the IQuantity object to use in the conversion, or the quantity type itself

        :type from_unit: L{FractionValue}
        :param from_unit:
            current unit of the given fraction value

        :type to_unit: L{FractionValue}
        :param to_unit:
            the unit to convert the fraction value to

        :rtype: L{FractionValue}
        :returns:
            the converted fraction value
        """
        # check if a quantity type unicode was passed
        if quantity.__class__ == six.text_type:
            # Note: actually ignoring the initial quantity type in this case because we
            # do the operation just using the from unit which may have any category (i.e.
            # the important thing is the quantity type, so, it can be created with the
            # default category).
            quantity = ObtainQuantity(from_unit)

        convert_to_quantity = ObtainQuantity(
            from_unit, quantity.GetComposingCategories()
        )
        converted_number = convert_to_quantity.ConvertScalarValue(
            fraction_value.GetNumber(), to_unit
        )

        # convert the number
        result = FractionValue(number=converted_number)
        # convert fraction's numerator
        if fraction_value.GetFraction() is not None:
            converted_numerator = convert_to_quantity.ConvertScalarValue(
                fraction_value.GetFraction().numerator, to_unit
            )

            converted_fraction = copy.copy(fraction_value.GetFraction())
            converted_fraction.numerator = converted_numerator
            result.SetFraction(converted_fraction)
        return result

    # GetFormatted ------------------------------------------------------------------------------------
    def GetFormattedValue(self, unit=None, value_format=None):
        """
        Returns the value part, that is, the number and fraction.

        :rtype: C{unicode}
        :returns:
            the formatted string
        """
        return self.GetValue(unit).GetLocalizedString()

    def __str__(self):
        """
        unicode() operator.

        :rtype: C{unicode}
        :returns:
            the string representation
        """
        return self.GetFormatted()

    if six.PY2:
        __unicode__ = __str__
        del __str__

    def GetFormatted(self, unit=None):
        """
        Returns the string representation for this FractionScalar.

        :param unicode unit:
            If None, returns the current unit, otherwise, returns the string representation of the
            value converted to the given unit.

        :rtype: unicode
        :returns:
            The string representation
        """
        return self.GetFormattedValue(unit) + self.GetFormattedSuffix(unit)

    def __repr__(self):
        """
        repr() operator.

        :rtype: C{unicode}
        :returns:
            the string representation
        """
        return "FractionScalar(%r, value=%r, unit=%r)" % (
            self._quantity.GetCategory(),
            self._value,
            self.unit,
        )

    # Equality -------------------------------------------------------------------------------------
    def __eq__(self, other):
        """
        == operator.

        :type other: L{FractionScalar}
        :param other:
            other fraction scalar

        :rtype: C{bool}
        :returns:
            if other is equal to self
        """
        return (
            type(self) == type(other)
            and self._value == other.value
            and self._quantity == other._quantity
        )

    def __lt__(self, other):
        """
        Comparison between objects.

        :param FractionScalar other:
            The object to be compared to

        :rtype: 1 if this object is greater than the other, -1 it is less than, 0 if they are equal.
        """
        # this is exactly the same comparison performed by the Scalar, however as they don't share
        # a base class where this method would fit, it was decided to implement it here, instead
        # of creating a base class just because of this method
        if self.quantity_type != other.quantity_type:
            msg = "can not compare scalars of different quantity types: %r != %r"
            raise TypeError(msg % self.quantity_type, other.quantity_type)

        v1 = self._value
        v2 = other.GetValue(self.unit)
        return v1 < v2

    # RegisterFractionScalarConversion -----------------------------------------
    @classmethod
    def RegisterFractionScalarConversion(cls):
        """
        Register a special unit conversion for FractionScalar.

        :param UnitDatabase db_unit:
            The unit-database instance to register the conversion into.
        """

        def ConvertFractionScalar(db_unit, quantity_type, from_unit, to_unit, value):
            """
            Converts the given Fraction Scalar by applying the converts method of this class.
            """
            converted = cls.ConvertFractionValue(
                value, quantity_type, from_unit, to_unit
            )
            return FractionValue(number=float(converted))

        UnitDatabase.RegisterAdditionalConversionType(
            FractionValue, ConvertFractionScalar
        )


FractionScalar.RegisterFractionScalarConversion()
